import re
import sys
from functools import partial
from typing import Iterator
from typing import List
from typing import Match
from typing import Optional

from jinja2 import contextfilter
from more_itertools import first
from unidecode import unidecode


def first_address_of_type(value: List[dict], address_type_uuid: str) -> Optional[str]:
    return first(
        (
            addr["value"]
            for addr in value
            if addr["address_type"]["uuid"] == address_type_uuid
        ),
        None,
    )


def location_element(value, index, sep="\\"):
    try:
        return value.split(sep)[index]
    except IndexError:
        return None


@contextfilter
def name_to_email_address(ctx, value):
    upn_end = ctx["_upn_end"]
    all_emails = ctx["_get_all_ad_emails"]()
    serial_start = 3

    def _split_and_normalize(name: str) -> List[str]:
        special = {ord("@"): None, ord("'"): None}

        parts: List[str] = re.split(r"[\-\s+]", name)
        parts = list(map(unidecode, parts))
        parts = list(map(str.lower, parts))
        parts = list(map(lambda s: s.translate(special), parts))

        if len(parts) < 2:
            raise ValueError("name must have at least two parts (name=%r)" % name)
        elif len(parts) == 2:
            return parts
        elif len(parts) > 2:
            return [parts[0], parts[1], parts[-1]]

        return []  # never reached, but makes mypy happy

    def _variant_a(parts: List[str]) -> str:
        # "firstname.lastname" if two name parts
        # "firstname.secondname.lastname" if more than two name parts
        return ".".join(parts)

    def _variant_b(parts: List[str]) -> str:
        # "f.lastname" if two name parts
        # "f.secondname.lastname" if more than two name parts
        return ".".join([parts[0][0]] + parts[1:])

    def _variant_c(parts: List[str]) -> str:
        # "firstname.l" if two name parts
        # "firstname.secondname.l" if more than two name parts
        return ".".join(parts[:-1] + [parts[-1][0]])

    def _variant_d(parts: List[str]) -> str:
        # Like variant A, except a serial number is appended.
        # If variants A, B, C and possibly D are already taken, find the next available
        # suffixed variant, e.g. "firstname.lastname.4" if "firstname.lastname.1" to
        # "firstname.lastname..3" are already taken.
        prefix = _variant_a(parts)

        def taken_suffixes() -> List[int]:
            # Look for emails matching e.g. "firstname.lastname.42@kommune.dk" and
            # extract the numeric part. Filter out non-matches.
            pattern: str = rf"{re.escape(prefix)}\.(\d+)@{re.escape(upn_end)}"
            prefix_matches: Iterator[Match] = filter(
                None, map(partial(re.search, pattern), all_emails)
            )
            # Extract numeric part, and cast to int.
            numeric_suffixes: Iterator[int] = map(
                lambda match: int(match.groups()[0]),
                prefix_matches,
            )
            return sorted(numeric_suffixes)

        def next_suffix():
            # Find next available suffix higher than or equal to `serial_start`.
            # E.g. if `taken_suffixes` is `{0, 1, 3, 4, 6}`, the next available suffix
            # is 5 - and not 2, as that would be below `serial_start` (which is 3.)
            taken = taken_suffixes()
            for new, curr in enumerate(taken + [sys.maxsize], serial_start):
                if new not in taken:
                    return new

        return f"{prefix:s}.{next_suffix():d}"

    def _gen_email(name: str) -> Optional[str]:
        parts = _split_and_normalize(name)

        for fn in (_variant_a, _variant_b, _variant_c, _variant_d):
            email = "%s@%s" % (fn(parts), upn_end)
            if email not in all_emails:
                return email

        raise ValueError("could not generate unique email address for %r" % value)

    return _gen_email(value)



from unittest import mock
from uuid import uuid4

import config
import los_import
from click.testing import CliRunner


class TestMain:
    def test_settings_exclude_empty_command_line_args(self):
        runner = CliRunner()
        with mock.patch.object(config.Settings, "from_kwargs") as mock_from_kwargs:
            # Run `main` as Click would
            runner.invoke(los_import.main)
            # Assert no (empty) Click options were passed to
            # `config.Settings.from_kwargs`.
            mock_from_kwargs.assert_called_once_with()

    def test_settings_include_command_line_args(self):
        runner = CliRunner()
        it_system_uuid = uuid4()  # Random valid UUID
        with mock.patch.object(config.Settings, "from_kwargs") as mock_from_kwargs:
            # Run `main` as Click would
            runner.invoke(
                los_import.main,
                ["--azid-it-system-uuid=%s" % it_system_uuid],
            )
            # Assert config was built, and that command line arg was used
            mock_from_kwargs.assert_called_once()
            kwargs = mock_from_kwargs.call_args.kwargs
            assert kwargs["azid_it_system_uuid"] == it_system_uuid



# SPDX-FileCopyrightText: Magenta ApS
#
# SPDX-License-Identifier: MPL-2.0
"""Integration endpoints."""
import logging
from threading import Lock

from fastapi import APIRouter
from fastapi import BackgroundTasks
from fastapi import HTTPException
from fastapi import Query
from fastramqpi.metrics import dipex_last_success_timestamp

from .config import DatabaseSettings
from .sql_export import SqlExport

logger = logging.getLogger(__name__)
trigger_router = APIRouter()
# Concurrency lock to ensure that only 1 operation is running at a time
lock_actual = Lock()
lock_historic = Lock()


def refresh_db(
    resolve_dar: bool, historic: bool, read_from_cache: bool, lock: Lock
) -> None:
    try:
        logger.info("*SQL export started*")
        database_settings = DatabaseSettings()

        sql_export = SqlExport(
            force_sqlite=False,
            historic=historic,
            settings=database_settings.to_old_settings(),
        )
        sql_export.perform_export(
            resolve_dar=resolve_dar,
            use_pickle=read_from_cache,
        )

        sql_export.swap_tables()
        logger.info("*SQL export ended*")
        dipex_last_success_timestamp.set_to_current_time()
    finally:
        # Another operation can start now
        lock.release()


@trigger_router.post("/trigger")
def trigger(
    background_tasks: BackgroundTasks,
    resolve_dar: bool = Query(False),
    historic: bool = Query(False),
    read_from_cache: bool = Query(False),
) -> dict[str, str]:
    if historic:
        lock = lock_historic
    else:
        lock = lock_actual
    acquired = lock.acquire(blocking=False)
    if not acquired:
        settings = DatabaseSettings()
        if settings.log_overlapping_aak:
            sql_export = SqlExport(
                force_sqlite=False,
                historic=historic,
                settings=settings.to_old_settings(),
            )
            sql_export.log_overlapping_runs_aak()
        raise HTTPException(409, "Already running")

    background_tasks.add_task(refresh_db, resolve_dar, historic, read_from_cache, lock)
    return {"detail": "Triggered"}



from datetime import datetime
from typing import Optional
from uuid import UUID


def create_employee(*, uuid: UUID, name, cpr_no):
    payload = {"type": "employee", "uuid": str(uuid), "name": name, "cpr_no": cpr_no}

    return payload


def create_engagement(
    *,
    uuid: UUID,
    org_unit_uuid: UUID,
    person_uuid: UUID,
    job_function_uuid: UUID,
    engagement_type_uuid: UUID,
    from_date: str,
    to_date: Optional[str],
    primary_uuid: UUID,
    user_key=str,
):
    payload = {
        "type": "engagement",
        "uuid": str(uuid),
        "org_unit": {"uuid": str(org_unit_uuid)},
        "person": {"uuid": str(person_uuid)},
        "job_function": {"uuid": str(job_function_uuid)},
        "primary": {"uuid": str(primary_uuid)},
        "engagement_type": {"uuid": str(engagement_type_uuid)},
        "user_key": user_key,
        "validity": {
            "from": from_date,
            "to": to_date,
        },
    }
    return payload


def create_org_unit(
    *,
    uuid: UUID,
    user_key: str,
    name: Optional[str],
    parent_uuid: UUID,
    org_unit_hierarchy: Optional[UUID],
    org_unit_type_uuid: Optional[UUID],
    from_date: str,
    to_date: Optional[str],
):
    payload = {
        "type": "org_unit",
        "uuid": str(uuid),
        "user_key": user_key,
        "parent": {"uuid": str(parent_uuid)},
        "validity": {"from": from_date, "to": to_date},
    }
    if name:
        payload["name"] = name
    if org_unit_hierarchy:
        payload["org_unit_hierarchy"] = {"uuid": str(org_unit_hierarchy)}
    if org_unit_type_uuid:
        payload["org_unit_type"] = {"uuid": str(org_unit_type_uuid)}
    return payload


def create_address(
    *,
    uuid: UUID,
    value: Optional[str],
    address_type_uuid: UUID,
    person_uuid: Optional[UUID] = None,
    org_unit_uuid: Optional[UUID] = None,
    from_date: str,
    to_date: Optional[str],
):
    payload = {
        "type": "address",
        "uuid": str(uuid),
        "value": value,
        "address_type": {"uuid": str(address_type_uuid)},
        "validity": {"from": from_date, "to": to_date},
    }
    if person_uuid:
        payload["person"] = {"uuid": str(person_uuid)}
    if org_unit_uuid:
        payload["org_unit"] = {"uuid": str(org_unit_uuid)}
    return payload


def create_it_rel(
    *,
    uuid: Optional[UUID],
    user_key: Optional[str],
    person_uuid: UUID,
    itsystem_uuid: Optional[UUID],
    from_date: str,
    to_date: Optional[str] = None,
):
    payload = {
        "type": "it",
        "uuid": str(uuid),
        "user_key": user_key,
        "person": {"uuid": str(person_uuid)},
        "itsystem": {"uuid": str(itsystem_uuid)},
        "validity": {"from": from_date, "to": to_date},
    }
    return payload


def create_manager(
    *,
    uuid: UUID,
    person_uuid: UUID,
    org_unit_uuid: UUID,
    manager_type_uuid: UUID,
    manager_level_uuid: UUID,
    responsibility_uuid: UUID,
    from_date: str,
    to_date: Optional[str] = None,
):
    payload = {
        "type": "manager",
        "uuid": str(uuid),
        "person": {"uuid": str(person_uuid)},
        "org_unit": {"uuid": str(org_unit_uuid)},
        "manager_type": {"uuid": str(manager_type_uuid)},
        "manager_level": {"uuid": str(manager_level_uuid)},
        "responsibility": [{"uuid": str(responsibility_uuid)}],
        "validity": {"from": from_date, "to": to_date},
    }

    return payload


def convert_create_to_edit(payload: dict, from_date: Optional[str] = None):
    """Convert an existing create payload to an edit payload"""
    edit_payload = {
        "data": {**payload},
        "uuid": payload["uuid"],
        "type": payload["type"],
    }
    if from_date:
        edit_payload["data"]["validity"] = {"from": from_date}

    return edit_payload


def terminate_detail(type: str, uuid: UUID, to_date: datetime):
    """Create payload for terminating a MO detail"""
    payload = {
        "type": type,
        "uuid": str(uuid),
        "validity": {"to": to_date.date().isoformat()},
    }
    return payload



from unittest import TestCase

import hypothesis.strategies as st
from hypothesis import given

from ..ad_reader import first_included

best_match = {
    "SamAccountName": "best_match",
    "xFieldname": "A",
}
match = {
    "SamAccountName": "match",
    "xFieldname": "C",
}
no_match = {
    "SamAccountName": "no_match",
    "xFieldname": "X",
}


class TestFirstIncluded(TestCase):
    def setUp(self):
        self.settings = {
            "discriminator.field": "xFieldname",
            "discriminator.values": ["A", "B", "C"],
            "discriminator.function": "include",
        }

    @given(st.permutations([best_match, match, no_match]))
    def test_first_included_user_order_irrelevant(self, users):
        result = first_included(self.settings, users)
        self.assertEqual(result, best_match)

    def test_substring_bug(self):
        # This users xFieldName is not even in discriminator.values
        # They should therefore not be included
        unwanted_match = {
            "SamAccountName": "unwanted_match",
            "xFieldname": "Adversary",
        }
        result = first_included(self.settings, [unwanted_match, match])
        self.assertEqual(result, match)



import logging
import sys

import click

from integrations.calculate_primary.common import LOGGER_NAME

# from click_option_group import RequiredMutuallyExclusiveOptionGroup, optgroup


def setup_logging():
    LOG_LEVEL = logging.DEBUG

    detail_logging = ("mora-helper", LOGGER_NAME)
    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        stream=sys.stdout,
    )


def get_engagement_updater(integration):
    if integration == "DEFAULT":
        from integrations.calculate_primary.default import (
            DefaultPrimaryEngagementUpdater,
        )

        return DefaultPrimaryEngagementUpdater
    if integration == "SD":
        from integrations.calculate_primary.sd import SDPrimaryEngagementUpdater

        return SDPrimaryEngagementUpdater
    if integration == "OPUS":
        from integrations.calculate_primary.opus import OPUSPrimaryEngagementUpdater

        return OPUSPrimaryEngagementUpdater
    raise NotImplementedError("Unexpected integration: " + str(integration))


@click.command()
@click.option(
    "--integration",
    type=click.Choice(["DEFAULT", "SD", "OPUS"], case_sensitive=False),
    required=True,
    help="Integration to use",
)
@click.option("--dry-run", is_flag=True, type=click.BOOL, help="Make no changes")
# @optgroup.group("Operation", cls=RequiredMutuallyExclusiveOptionGroup, help="")
# @optgroup.option(
#    "--check-all", is_flag=True, type=click.BOOL, help="Check all users"
# )
# @optgroup.option("--check-user", type=click.UUID, help="Check one user")
# @optgroup.option(
#    "--recalculate-all", is_flag=True, type=click.BOOL, help="Recalculate all users"
# )
# @optgroup.option("--recalculate-user", type=click.UUID, help="Recalculate one user")
@click.option("--check-all", is_flag=True, type=click.BOOL, help="Check all users")
@click.option("--check-user", type=click.UUID, help="Check one user")
@click.option(
    "--recalculate-all", is_flag=True, type=click.BOOL, help="Recalculate all users"
)
@click.option("--recalculate-user", type=click.UUID, help="Recalculate one user")
def calculate_primary(
    integration, dry_run, check_all, check_user, recalculate_all, recalculate_user
):
    """Tool to work with primary engagement(s)."""
    setup_logging()

    # Acquire the configured updater
    updater_class = get_engagement_updater(integration)
    updater = updater_class(dry_run=dry_run)

    # Run the specified operation
    if check_all:
        print("Check all")
        updater.check_all()
    if check_user:
        print("Check user")
        updater.check_user(check_user)
    if recalculate_all:
        print("Recalculate all")
        updater.recalculate_all()
    if recalculate_user:
        print("Recalculate user")
        print(updater.recalculate_user(recalculate_user))


if __name__ == "__main__":
    calculate_primary()  # type: ignore



import datetime

from integrations.calculate_primary.common import logger
from integrations.calculate_primary.common import MOPrimaryEngagementUpdater


# This function has been copied directly from the SD-integration repo as
# it was missing when the SD-integration was moved to its own repo. Ideally,
# it should have been moved to a library and used both in DIPEX and in the
# SD-integration, but it is probably not worth the effort, since it is not
# unlikely that it will be removed from the SD-integration anyway as this
# integration should not have the responsibility of calculating primary
# engagements
def get_primary_types(helper):
    """
    Read the engagement types from MO and match them up against the four
    known types in the SD->MO import.

    Args:
        helper: An instance of mora-helpers.

    Returns:
        A dict matching up the engagement types with LoRa class uuids (
        i.e. UUIDs of facets).

    Example:
        An example return value:
        ```python
        {
            "primary": "697c8838-ba0f-4e74-90f8-4e7c31d4e7e7",
            "non_primary": "b9543f90-9511-494b-bbf5-f15678502c2d",
            "no_salary": "88589e84-5736-4f8c-9c0c-2e29046d7471",
            "fixed_primary": "c95a1999-9f95-4458-a218-e9c96e7ad3db",
        }
        ```
    """

    # These constants are global in all SD municipalities (because they are created
    # by the SD->MO importer.
    PRIMARY = "Ansat"
    NO_SALARY = "status0"
    NON_PRIMARY = "non-primary"
    FIXED_PRIMARY = "explicitly-primary"

    logger.info("Read primary types")
    primary = None
    no_salary = None
    non_primary = None
    fixed_primary = None

    primary_types = helper.read_classes_in_facet("primary_type")
    for primary_type in primary_types[0]:
        if primary_type["user_key"] == PRIMARY:
            primary = primary_type["uuid"]
        if primary_type["user_key"] == NON_PRIMARY:
            non_primary = primary_type["uuid"]
        if primary_type["user_key"] == NO_SALARY:
            no_salary = primary_type["uuid"]
        if primary_type["user_key"] == FIXED_PRIMARY:
            fixed_primary = primary_type["uuid"]

    type_uuids = {
        "primary": primary,
        "non_primary": non_primary,
        "no_salary": no_salary,
        "fixed_primary": fixed_primary,
    }
    if None in type_uuids.values():
        raise Exception("Missing primary types: {}".format(type_uuids))
    return type_uuids


class SDPrimaryEngagementUpdater(MOPrimaryEngagementUpdater):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def remove_past(user_uuid, no_past, eng):
            if no_past and eng["validity"]["to"]:
                to = datetime.datetime.strptime(eng["validity"]["to"], "%Y-%m-%d")
                if to < datetime.datetime.now():
                    return False
            return True

        def remove_no_salary(eng):
            return not self._predicate_primary_is("no_salary", eng)

        def remove_no_salary_check(user_uuid, eng):
            return remove_no_salary(eng)

        def remove_no_salary_calculate(user_uuid, no_past, eng):
            return remove_no_salary(eng)

        def remove_missing_user_key(user_uuid, no_past, eng):
            return "user_key" in eng

        self.check_filters = [
            remove_no_salary_check,
        ]

        self.calculate_filters = [
            remove_past,
            remove_no_salary_calculate,
            remove_missing_user_key,
        ]

    def _find_primary_types(self):
        # Keys are; fixed_primary, primary, no_salary and non-primary
        primary_types = get_primary_types(self.helper)
        primary = [
            primary_types["fixed_primary"],
            primary_types["primary"],
            primary_types["no_salary"],
        ]
        return primary_types, primary

    def _find_primary(self, mo_engagements):
        def non_integer_userkey(mo_engagement):
            try:
                # non-integer user keys should universally be status0, and as such
                # they should already have been filtered out, thus if they have not
                # been filtered out, they must have the wrong primary_type.
                int(mo_engagement["user_key"])
            except ValueError:
                self._fixup_status_0(mo_engagement)
                # Filter it out, as it should have been
                return False
            return True

        # Ensure that all mo_engagements have integer user_keys.
        mo_engagements = list(filter(non_integer_userkey, mo_engagements))

        if not mo_engagements:
            return None

        # The primary engagement is the engagement with the highest occupation rate.
        # - The occupation rate is found as 'fraction' on the engagement.
        #
        # If two engagements have the same occupation rate, the tie is broken by
        # picking the one with the lowest user-key integer.
        primary_engagement = max(
            mo_engagements,
            # Sort first by fraction, then reversely by user_key integer
            key=lambda eng: (eng.get("fraction") or 0, -int(eng["user_key"])),
        )
        return primary_engagement["uuid"]

    def _fixup_status_0(self, mo_engagement):
        logger.warning("Engagement type not status0. Will fix.")
        validity = mo_engagement["validity"]
        payload = {
            "type": "engagement",
            "uuid": mo_engagement["uuid"],
            "data": {
                "primary": {"uuid": self.primary_types["no_salary"]},
                "validity": validity,
            },
        }
        logger.debug("Status0 edit payload: {}".format(payload))
        if not self.dry_run:
            response = self.helper._mo_post("details/edit", payload)
            assert response.status_code == 200
            logger.info("Status0 fixed")



import datetime
import hashlib
import logging
import pathlib
import re
import sqlite3
import uuid
from functools import lru_cache
from operator import itemgetter
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import xmltodict
from deepdiff import DeepDiff
from more_itertools import first
from more_itertools import partition
from ra_utils.load_settings import load_settings
from ra_utils.tqdm_wrapper import tqdm

from integrations import cpr_mapper
from integrations.opus.opus_exceptions import ImporterrunNotCompleted
from integrations.opus.opus_file_reader import get_opus_filereader

SETTINGS = load_settings()
START_DATE = datetime.datetime(2019, 1, 1, 0, 0)

logger = logging.getLogger("opusHelper")


def read_cpr_mapping():
    cpr_map = pathlib.Path.cwd() / "settings" / "cpr_uuid_map.csv"
    if not cpr_map.is_file():
        logger.error("Did not find cpr mapping")
        raise Exception("Did not find cpr mapping")

    logger.info("Found cpr mapping")
    employee_forced_uuids = cpr_mapper.employee_mapper(str(cpr_map))
    return employee_forced_uuids


def read_available_dumps() -> Dict[datetime.datetime, str]:
    dumps = get_opus_filereader().list_opus_files()
    assert len(dumps) > 0, "No Opus files found!"
    return dumps


def get_latest_dump():
    dumps = read_available_dumps()
    latest_date = max(dumps.keys())
    return latest_date, dumps[latest_date]


def local_db_insert(insert_tuple):
    conn = sqlite3.connect(
        SETTINGS["integrations.opus.import.run_db"],
        detect_types=sqlite3.PARSE_DECLTYPES,
    )
    c = conn.cursor()
    query = "insert into runs (dump_date, status) values (?, ?)"
    final_tuple = (insert_tuple[0], insert_tuple[1].format(datetime.datetime.now()))
    c.execute(query, final_tuple)
    conn.commit()
    conn.close()


def initialize_db(run_db):
    logger.info("Force is true, create new db")
    conn = sqlite3.connect(str(run_db))
    c = conn.cursor()
    c.execute(
        """
    CREATE TABLE runs (id INTEGER PRIMARY KEY,
    dump_date timestamp, status text)
    """
    )
    conn.commit()
    conn.close()


def next_xml_file(run_db, dumps) -> Tuple[Optional[datetime.date], datetime.date]:
    conn = sqlite3.connect(
        SETTINGS["integrations.opus.import.run_db"],
        detect_types=sqlite3.PARSE_DECLTYPES,
    )
    c = conn.cursor()
    query = "select * from runs order by id desc limit 1"
    c.execute(query)
    row = c.fetchone()
    latest_date = row[1]
    next_date = None
    if "Running" in row[2]:
        print("Critical error")
        logging.error("Previous run did not return!")
        raise ImporterrunNotCompleted("Previous run did not return!")

    for date in sorted(dumps.keys()):
        if date > latest_date:
            next_date = date
            break

    return next_date, latest_date


def parse_phone(phone_number):
    validated_phone = None
    if len(phone_number) == 8:
        validated_phone = phone_number
    elif len(phone_number) in (9, 11):
        validated_phone = phone_number.replace(" ", "")
    elif len(phone_number) in (4, 5):
        validated_phone = "0000" + phone_number.replace(" ", "")

    if validated_phone is None:
        logger.warning("Could not parse phone {}".format(phone_number))
    return validated_phone


@lru_cache(maxsize=None)
def generate_uuid(value):
    """
    Generate a predictable uuid based on org name and a unique value.
    """
    base_hash = hashlib.md5(SETTINGS["municipality.name"].encode())
    base_digest = base_hash.hexdigest()
    base_uuid = uuid.UUID(base_digest)

    combined_value = (str(base_uuid) + str(value)).encode()
    value_hash = hashlib.md5(combined_value)
    value_digest = value_hash.hexdigest()
    value_uuid = uuid.UUID(value_digest)
    return value_uuid


def gen_unit_uuid(unit):
    """generate uuids for given units."""
    return str(generate_uuid(unit["@id"]))


def parser(target_file: str, opus_id: Optional[int] = None) -> Tuple[List, List]:
    """Read an opus file and return units and employees"""
    text_input = get_opus_filereader().read_file(target_file)

    data = xmltodict.parse(text_input)
    data = data["kmd"]
    units = data.get("orgUnit", [])
    employees = data.get("employee", [])
    if opus_id is not None:
        employees = list(filter(lambda x: int(x["@id"]) == opus_id, employees))
        units = list(filter(lambda x: int(x["@id"]) == opus_id, units))
    return units, employees


def find_changes(
    before: List[Dict], after: List[Dict], disable_tqdm: bool = True
) -> List[Dict]:
    """Filter a list of dictionaries based on differences to another list of dictionaries
    Used to find changes to org_units and employees in opus files.
    Any registration in lastChanged is ignored here.
    Use disable_tqdm in tests etc.

    Returns: list of dictionaries from 'after' where there are changes from 'before'
    >>> a = [{"@id":1, "text":"unchanged", '@lastChanged': 'some day'}, {"@id":2, "text":"before", '@lastChanged':'today'}]
    >>> b = [{"@id":1, "text":"unchanged", '@lastChanged': 'another day'}, {"@id":2, "text":"after"}]
    >>> c = [{"@id":1, "text":"unchanged", '@lastChanged': 'another day'}]
    >>> find_changes(a, a, disable_tqdm=True)
    []
    >>> find_changes(a, b, disable_tqdm=True)
    [{'@id': 2, 'text': 'after'}]
    >>> find_changes(b, a, disable_tqdm=True)
    [{'@id': 2, 'text': 'before', '@lastChanged': 'today'}]
    >>> find_changes(a, c, disable_tqdm=True)
    []
    """
    old_ids = list(map(itemgetter("@id"), before))
    old_map = dict(zip(old_ids, before))
    changed_obj = []

    def find_changed(obj: Dict) -> bool:
        # New object
        if obj["@id"] not in old_ids:
            return True

        old_obj = old_map[obj["@id"]]
        diff = DeepDiff(
            obj,
            old_obj,
            exclude_paths={
                "root['@lastChanged']",
                "root['numerator']",
                "root['denominator']",
            },
        )
        # Changed object
        if diff:
            return True

        # Unchanged object
        return False

    after = tqdm(after, desc="Finding changes", disable=disable_tqdm)
    changed_obj = list(filter(find_changed, after))

    return changed_obj


def find_missing(before: List[Dict], after: List[Dict]) -> List[Dict]:
    """Check if an element is missing. This happens when an object is cancled in Opus.

    >>> a = [{"@id":1}, {"@id":2}, {"@id":3}]
    >>> b = [{"@id":1}, {"@id":3}]
    >>> find_missing(a,b)
    [{'@id': 2}]
    >>> find_missing(b,a)
    []
    """

    old_ids = set(map(itemgetter("@id"), before))
    new_ids = set(map(itemgetter("@id"), after))
    missing = old_ids - new_ids
    missing_elements = filter(lambda x: x.get("@id") in missing, before)
    return list(missing_elements)


def file_diff(
    file1: Optional[str],
    file2: str,
    disable_tqdm: bool = True,
    opus_id: Optional[int] = None,
):
    """Compares two files and returns all units and employees that have been changed."""
    units1: List[Dict] = []
    employees1: List[Dict] = []
    if file1:
        units1, employees1 = parser(file1, opus_id=opus_id)
    units2, employees2 = parser(file2, opus_id=opus_id)

    units = find_changes(units1, units2, disable_tqdm=disable_tqdm)
    cancelled_units = find_missing(units1, units2)

    employees = find_changes(employees1, employees2, disable_tqdm=disable_tqdm)
    cancelled_employees = find_missing(employees1, employees2)

    return {
        "units": units,
        "employees": employees,
        "cancelled_units": cancelled_units,
        "cancelled_employees": cancelled_employees,
    }


def compare_employees(original, new, force=False):
    """Differences is these keys will not be counted as a difference, unless force
    is set to true. Notice lastChanged is included here, since we perform a
    brute-force comparison and does not care for lastChanged.
    """
    skip_keys = [
        "productionNumber",
        "entryIntoGroup",
        "invoiceRecipient",
        "@lastChanged",
        "cpr",
    ]
    identical = True
    for key in new.keys():
        if key in skip_keys and not force:
            continue
        if not original.get(key) == new[key]:
            identical = False
            msg = "Changed {} from {} to {}"
            print(msg.format(key, original.get(key), new[key]))
    return identical


def filter_units(units, filter_ids):
    """Splits units into two based on filter_ids.

    Partitions the units such that no unit with a parent-id in filter_ids exist in one list.
    Any unit filtered like that is put in the other list.

    Example:
        >>> units = [(1, None), (2, 1), (3, 1), (4, 2), (5, 2), (6, 3), (7, 5)]
        >>> tup_to_unit = lambda tup: {'@id': tup[0], 'parentOrgUnit': tup[1]}
        >>> units = list(map(tup_to_unit, units))
        >>> get_ids = lambda units: list(map(itemgetter('@id'), units))
        >>> a, b = filter_units(units, [1])
        >>> get_ids(a)
        [1, 2, 3, 4, 5, 6, 7]
        >>> get_ids(b)
        []
        >>> a, b = filter_units(units, [2])
        >>> get_ids(a)
        [2, 4, 5, 7]
        >>> get_ids(b)
        [1, 3, 6]
        >>> a, b = filter_units(units, [3])
        >>> get_ids(a)
        [3, 6]
        >>> get_ids(b)
        [1, 2, 4, 5, 7]
        >>> a, b = filter_units(units, [3, 5])
        >>> get_ids(a)
        [3, 5, 6, 7]
        >>> get_ids(b)
        [1, 2, 4]
        >>> a, b = filter_units(units, [3, 7])
        >>> get_ids(a)
        [3, 6, 7]
        >>> get_ids(b)
        [1, 2, 4, 5]

    Args:
        units: List of units
        filter_ids: List of unit IDs to filter parents on

    Returns:
        list: List of units, with some filtered out
    """

    def get_parent(parent_map, entry):
        """Build a list of parents."""
        parent = parent_map.get(entry, None)
        if parent is None:
            return [entry]
        return [entry] + get_parent(parent_map, parent)

    parent_map = dict(map(itemgetter("@id", "parentOrgUnit"), units))
    filter_set = set(filter_ids)

    def is_disjoint_from_filter_ids(unit):
        """Test for overlap between parents and filter_set."""
        parent_set = set(get_parent(parent_map, unit["@id"]))
        return parent_set.isdisjoint(filter_set)

    return partition(is_disjoint_from_filter_ids, units)


def filter_employees(employees: Iterable[Dict], all_filtered_ids: set):
    """Remove any employees that has an engagement in an unit that is in all_filtered_ids

    >>> e = [{'orgUnit': "1"}, {'orgUnit': "2"}]
    >>> ids = {"2", "3"}
    >>> list(filter_employees(e, ids))
    [{'orgUnit': '1'}]

    """
    return filter(lambda empl: empl.get("orgUnit") not in all_filtered_ids, employees)


def split_employees_leaves(employees: List[Dict]) -> Tuple[Iterable, Iterable]:
    """Split list of employees into two iterables, with either active employees or terminated employees

    >>> e = [{'@action': "test"}, {'@action': "leave"}]
    >>> e1, e2 = split_employees_leaves(e)
    >>> list(e1)
    [{'@action': 'test'}]

    >>> list(e2)
    [{'@action': 'leave'}]
    """

    return partition(lambda empl: empl.get("@action") == "leave", employees)


def read_cpr(employee: dict) -> str:
    cpr = employee.get("cpr")
    if isinstance(cpr, dict):
        cpr = employee["cpr"]["#text"]
    elif isinstance(cpr, str):
        assert isinstance(int(cpr), int)
    else:
        raise TypeError("Can't read cpr in this format")
    return cpr


def find_all_filtered_units(inputfile, filter_ids) -> list[dict]:
    file_diffs = file_diff(None, inputfile)
    all_units = file_diffs["units"]
    all_units.extend(file_diffs["cancelled_units"])
    all_filtered_units, _ = filter_units(all_units, filter_ids)
    return list(all_filtered_units)


def include_cancelled(filename: str, employees, cancelled_employees) -> List:
    """Add cancelled employees to employees list, but set leavedate to date from filename

    >>> include_cancelled('./ZLPE202001010253_delta.xml', [], [{"id":1}])
    [{'id': 1, 'leaveDate': '2020-01-01'}]
    """

    filedate = re.search(r"\d{8}", str(filename))  # type: ignore
    filedatetime = datetime.datetime.strptime(filedate.group(), "%Y%m%d")  # type: ignore
    filedatestr = filedatetime.strftime("%Y-%m-%d")
    for empl in cancelled_employees:
        empl["leaveDate"] = filedatestr
    employees.extend(cancelled_employees)
    return employees


def read_and_transform_data(
    inputfile1: Optional[str],
    inputfile2: str,
    filter_ids: List[str],
    disable_tqdm=False,
    opus_id: Optional[int] = None,
) -> Tuple[Iterable, Iterable, Iterable, Iterable]:
    """Gets the diff of two files and transporms the data based on filter_ids
    Returns the active units, filtered units, active employees which are not in a filtered unit and employees which are terminated
    """
    file_diffs = file_diff(
        inputfile1, inputfile2, disable_tqdm=disable_tqdm, opus_id=opus_id
    )

    employees = include_cancelled(
        inputfile2, file_diffs["employees"], file_diffs["cancelled_employees"]
    )
    all_filtered_units = find_all_filtered_units(inputfile2, filter_ids)
    _, units = filter_units(file_diffs["units"], filter_ids)
    active_employees, terminated_employees = split_employees_leaves(employees)
    filtered_employees = filter_employees(
        active_employees, {unit["@id"] for unit in all_filtered_units}
    )
    return (
        list(units),
        list(all_filtered_units),
        list(filtered_employees),
        list(terminated_employees),
    )


@lru_cache
def find_opus_root_unit_uuid() -> uuid.UUID:
    """Generates uuid for opus root.

    Reads the first available opus file and generates the uuid for the first unit in the file.
    Assumes this is the root organisation of opus.
    """
    dumps = read_available_dumps()

    first_date = min(sorted(dumps.keys()))
    units, _ = parser(dumps[first_date])
    main_unit = first(units)
    calculated_uuid = generate_uuid(main_unit["@id"])
    return calculated_uuid



from typing import Optional


def create_user(cpr: str, ad_user: dict, org_uuid, uuid: Optional[str] = None) -> dict:
    payload = {
        "cpr_no": cpr,
        "uuid": uuid or ad_user["ObjectGUID"],
        "givenname": ad_user["GivenName"],
        "surname": ad_user["Surname"],
        "org": {"uuid": org_uuid},
    }
    return payload


def connect_it_system_to_user(user_uuid, username, it_system, from_date="1930-01-01"):
    payload = {
        "type": "it",
        "user_key": username,
        "itsystem": {"uuid": it_system},
        "person": {"uuid": user_uuid},
        "validity": {"from": from_date, "to": None},
    }
    return payload


# almost all parameters could be replaced with direct reading from settings
def create_engagement(
    ad_user, unit_uuid, person_uuid, job_function, engagement_type, validity
):
    payload = {
        "type": "engagement",
        "org_unit": {"uuid": str(unit_uuid)},
        "person": {"uuid": str(person_uuid)},
        "job_function": {"uuid": job_function},
        "engagement_type": {"uuid": engagement_type},
        "user_key": ad_user["SamAccountName"],
        "validity": validity,
    }
    return payload


def create_unit(uuid, unit_name, unit_type, parent):
    payload = {
        "uuid": uuid,
        "user_key": unit_name,
        "name": unit_name,
        "parent": {"uuid": parent},
        "org_unit_type": {"uuid": unit_type},
        "validity": {"from": "1930-01-01", "to": None},
    }
    return payload


def terminate_engagement(uuid, terminate_date):
    payload = {
        "type": "engagement",
        "uuid": uuid,
        "validity": {"to": terminate_date.strftime("%Y-%m-%d")},
    }
    return payload


# Same code is found in sd_payloads and opus_payloads
def klasse(bvn, navn, org, facet_uuid):
    validity = {"from": "1930-01-01", "to": "infinity"}

    properties = {
        "brugervendtnoegle": bvn,
        "titel": navn,
        "omfang": "TEXT",
        "virkning": validity,
    }
    attributter = {"klasseegenskaber": [properties]}
    relationer = {
        "ansvarlig": [
            {"objekttype": "organisation", "uuid": org, "virkning": validity}
        ],
        "facet": [{"objekttype": "facet", "uuid": facet_uuid, "virkning": validity}],
    }
    tilstande = {
        "klassepubliceret": [{"publiceret": "Publiceret", "virkning": validity}]
    }

    payload = {
        "attributter": attributter,
        "relationer": relationer,
        "tilstande": tilstande,
    }
    return payload



from uuid import uuid4

import config
import pytest
import uuids

from .helpers import HelperMixin


class TestConfig(HelperMixin):
    def test_get_config_fails_if_unconfigured(self):
        with pytest.raises(config.ImproperlyConfigured):
            config.get_config()

    def test_get_config_succeeds_if_configured(self):
        with self._mock_settings_json():
            settings_a = config.Settings.from_kwargs(**self._get_config_kwargs())
            settings_b = config.get_config()
            assert settings_a == settings_b

    def test_settings_include_default(self):
        with self._mock_settings_json():
            settings = config.Settings.from_kwargs(**self._get_config_kwargs())
            assert settings.azid_it_system_uuid == uuids.AZID_SYSTEM

    def test_settings_include_json(self):
        azid_uuid = uuid4()
        settings = {"integrations.aarhus_los.azid_it_system_uuid": str(azid_uuid)}
        with self._mock_settings_json(settings):
            settings = config.Settings.from_kwargs(**self._get_config_kwargs())
            assert settings.azid_it_system_uuid == azid_uuid

    def _get_config_kwargs(self):
        # Provide values for all settings that do not have a default value
        return dict(
            ftp_url="ftp-url",
            ftp_user="ftp-user",
            ftp_pass="ftp-pass",
            ftp_folder="ftp-folder",
            import_state_file="import-state-file",
        )



# -*- coding: utf-8 -*-
import copy
import json
import logging
import os
import random
import re
import time
from abc import ABC
from abc import abstractmethod
from datetime import date
from datetime import datetime
from functools import lru_cache
from functools import partial
from operator import itemgetter

import click
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup
from gql import gql
from jinja2 import Environment
from jinja2 import StrictUndefined
from jinja2 import Undefined
from more_itertools import first
from more_itertools import unzip
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.lazy_dict import LazyDict
from ra_utils.lazy_dict import LazyEval
from ra_utils.lazy_dict import LazyEvalDerived
from raclients.graph.client import GraphQLClient

from . import ad_templates
from .ad_common import AD
from .ad_exceptions import CprNotFoundInADException
from .ad_exceptions import CprNotNotUnique
from .ad_exceptions import EngagementDatesError
from .ad_exceptions import NoActiveEngagementsException
from .ad_exceptions import NoPrimaryEngagementException
from .ad_exceptions import ReplicationFailedException
from .ad_exceptions import SamAccountNameNotUnique
from .ad_exceptions import UserNotFoundException
from .ad_jinja_filters import first_address_of_type
from .ad_jinja_filters import location_element
from .ad_jinja_filters import name_to_email_address
from .ad_logger import start_logging
from .ad_reader import ADParameterReader
from .ad_template_engine import INVALID
from .ad_template_engine import prepare_field_templates
from .ad_template_engine import template_powershell
from .user_names import UserNameGen
from .utils import dict_exclude
from .utils import dict_map
from .utils import dict_subset
from .utils import lower_list


logger = logging.getLogger("AdWriter")


class MODataSource(ABC):
    @abstractmethod
    def read_user(self, uuid):
        """Read a user from MO using the provided uuid.

        Throws UserNotFoundException if the user cannot be found.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict: A dict with the users data.
        """
        raise NotImplementedError

    @abstractmethod
    def get_email_address(self, uuid):
        """Read a users email address using the provided uuid.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict: A dict of email address, potentially empty.
        """
        raise NotImplementedError

    @abstractmethod
    def find_primary_engagement(self, uuid):
        """Find the primary engagement for the provided uuid user.

        Args:
            uuid: UUID for the user to find primary engagement for.

        Returns:
            tuple(string, string, string, string):
                employment_number: Identifier for the engagement.
                title: Title of the job function for the engagement
                eng_org_unit: UUID of the organisation unit for the engagement
                eng_uuid: UUID of the found engagement
        """
        raise NotImplementedError

    @abstractmethod
    def get_manager_uuid(self, mo_user, eng_uuid):
        """Get UUID of the relevant manager for the user.

        Args:
            mo_user: MO user object, as returned by read_user.
            eng_uuid: UUID of the engagement, as returned by find_primary_engagement.

        Returns:
            str: A UUID string for the manager
        """
        raise NotImplementedError

    @abstractmethod
    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        raise NotImplementedError

    def get_engagement_endpoint_dates(self, uuid):
        """Return the earliest start- and latest end-date for the users engagements.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[str, str]: A tuple with start and end date.
        """
        start_dates, end_dates = self.get_engagement_dates(uuid)

        start_dates = map(lambda date: date if date else "1930-01-01", start_dates)
        start_date = min(start_dates, default="9999-12-31")

        end_dates = map(lambda date: date if date else "9999-12-31", end_dates)
        end_date = max(end_dates, default="1930-01-01")

        if date.fromisoformat(start_date) > date.fromisoformat(end_date):
            raise EngagementDatesError(
                "Invalid Engagement dates interval. "
                f"Start date {start_date} is greater than end date {end_date}"
            )

        return start_date, end_date

    @abstractmethod
    def get_it_systems(self, uuid):
        """Read the IT system bindings from the user using the provided uuid.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            dict of dicts: A dictionary it-system uuid to it-system dictionaries.
        """
        raise NotImplementedError


class LoraCacheSource(MODataSource):
    """LoraCache implementation of the MODataSource interface."""

    def __init__(self, lc, lc_historic, mo_rest_source):
        self.lc = lc
        self.lc_historic = lc_historic
        self.mo_rest_source = mo_rest_source

    def read_user(self, uuid):
        if uuid not in self.lc.users:
            raise UserNotFoundException()

        lc_user = self.lc.users[uuid][0]
        mo_user = {
            "uuid": uuid,
            "name": lc_user["navn"],
            "surname": lc_user["efternavn"],
            "givenname": lc_user["fornavn"],
            "nickname": lc_user["kaldenavn"],
            "nickname_givenname": lc_user["kaldenavn_fornavn"],
            "nickname_surname": lc_user["kaldenavn_efternavn"],
            "cpr_no": lc_user["cpr"],
        }
        return mo_user

    def get_email_address(self, uuid):
        mail_dict = {}
        for addr in self.lc.addresses.values():
            if addr[0]["user"] == uuid and addr[0]["scope"] == "E-mail":
                mail_dict = addr[0]
        return dict_subset(mail_dict, ["uuid", "value"])

    def find_primary_engagement(self, uuid):
        def filter_for_user(engagements):
            return filter(lambda eng: eng[0]["user"] == uuid, engagements)

        def filter_primary(engagements):
            return filter(lambda eng: eng[0]["primary_boolean"], engagements)

        user_engagements = list(filter_for_user(self.lc.engagements.values()))
        # No user engagements
        if not user_engagements:
            # But we may still have future engagements
            future_engagement = next(
                filter_for_user(self.lc_historic.engagements.values()), None
            )
            # We do not have any engagements at all
            if future_engagement is None:
                raise NoActiveEngagementsException()
            # We have future engagements, but LoraCache does not handle that.
            # Delegate to MORESTSource
            logger.info("Found future engagement")
            return self.mo_rest_source.find_primary_engagement(uuid)

        primary_engagement = next(filter_primary(user_engagements), None)
        if primary_engagement is None:
            raise NoPrimaryEngagementException("User: {}".format(uuid))

        primary_engagement = primary_engagement[0]
        employment_number = primary_engagement["user_key"]
        title = self.lc.classes[primary_engagement["job_function"]]["title"]
        eng_org_unit = primary_engagement["unit"]
        eng_uuid = primary_engagement["uuid"]
        return employment_number, title, eng_org_unit, eng_uuid

    def get_manager_uuid(self, mo_user, eng_uuid):
        def org_uuid_parent(org_uuid):
            parent_uuid = self.lc.units[org_uuid][0]["parent"]
            return parent_uuid

        def org_uuid_to_manager(org_uuid):
            org_unit = self.lc.units[org_uuid][0]
            manager_uuid = self.lc.managers[org_unit["acting_manager_uuid"]][0]["user"]
            return manager_uuid

        try:
            # Compatibility to mimic MORESTSource behaviour
            # MORESTSource does an engagement lookup in the present, using
            # the org uuid from that and fails if it doesn't find anything
            engagement = self.lc.engagements[eng_uuid][0]
            eng_org_unit = engagement["unit"]
            manager_uuid = org_uuid_to_manager(eng_org_unit)
            if manager_uuid is None:
                raise Exception("Unable to find manager")
            # We found a manager directly
            if manager_uuid != mo_user["uuid"]:
                return manager_uuid
            # Self manager, find a manager above us, if possible
            parent_uuid = org_uuid_parent(eng_org_unit)
            while manager_uuid == mo_user["uuid"]:
                if parent_uuid is None:
                    return manager_uuid
                manager_uuid = org_uuid_to_manager(parent_uuid)
                parent_uuid = org_uuid_parent(parent_uuid)
            return manager_uuid
        except KeyError:
            return None

    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        return self.mo_rest_source.get_engagement_dates(uuid)

    def get_it_systems(self, uuid):
        user_itsystems = filter(
            lambda eng: eng["user"] == uuid,
            map(itemgetter(0), self.lc.it_connections.values()),
        )
        return {it_system["itsystem"]: it_system for it_system in user_itsystems}


class MOGraphqlSource:
    """MO Graphql."""

    def __init__(self, settings):
        self._settings = settings
        self._response = self._run_query()
        self._manager_map = self._create_manager_map()

    def _get_client(self):
        return GraphQLClient(
            url=f"{self._settings['global']['mora.base']}/graphql/v7",
            client_id=os.environ.get("CLIENT_ID", "dipex"),
            client_secret=os.environ["CLIENT_SECRET"],
            auth_realm=os.environ.get("AUTH_REALM", "mo"),
            auth_server=os.environ["AUTH_SERVER"],
            sync=True,
            httpx_client_kwargs={"timeout": None},
        )

    def _run_query(self):
        query = gql(
            """
            query PaginatedOrgUnits($cursor: Cursor) {
              org_units(to_date: null, limit: 50, cursor: $cursor) {
                page_info {
                  next_cursor
                }
                objects {
                  objects {
                    uuid
                    parent_uuid
                    engagements {
                      employee_uuid
                      is_primary
                      validity {
                        to
                      }
                    }
                    managers(inherit: true) {
                      employee_uuid
                    }
                  }
                }
              }
            }
            """
        )
        result = []
        cursor = "initial"
        with self._get_client() as session:
            while cursor is not None:
                response = session.execute(
                    query,
                    variable_values={"cursor": None if cursor == "initial" else cursor},
                )
                cursor = response["org_units"]["page_info"]["next_cursor"]
                result.extend(response["org_units"]["objects"])
        return result

    def _create_manager_map(self):
        """Create mapping between employees and managers"""

        # Step 1: build intermediate dictionaries `engagement_map` and `org_unit_map`
        # from the GraphQL objects in `self._response`.

        # `engagement_map` maps an employee UUID to the org unit of the primary
        # engagement.
        engagement_map = {}

        # `org_unit_map` maps an org unit UUID to its parent org unit UUID as well as
        # its manager UUID.
        org_unit_map = {}

        objs = (obj[0] for obj in (item["objects"] for item in self._response))
        for obj in objs:
            # Skip org units without managers
            if not obj["managers"]:
                continue
            org_unit_uuid = obj["uuid"]
            parent_org_unit_uuid = obj["parent_uuid"]
            manager_uuid = first(obj["managers"])["employee_uuid"]
            # Populate `org_unit_map`
            org_unit_map[org_unit_uuid] = {
                "parent_uuid": parent_org_unit_uuid,
                "manager_uuid": manager_uuid,
            }
            # Populate `engagement_map`
            for engagement in obj["engagements"]:
                if engagement["is_primary"]:
                    employee_uuid = engagement["employee_uuid"]
                    engagement_map[employee_uuid] = org_unit_uuid

        # Step 2: visit each employee UUID in `engagement_map` and find its manager UUID
        # either by taking it directly from `org_unit_map`, or by traversing "up" the
        # org unit hierarchy, until the manager UUID differs from the employee UUID.
        result = {}
        for employee_uuid in engagement_map:
            org_unit_uuid = engagement_map[employee_uuid]
            manager_uuid = org_unit_map[org_unit_uuid]["manager_uuid"]
            while manager_uuid == employee_uuid:
                # Go to parent org unit UUID
                org_unit_uuid = org_unit_map[org_unit_uuid]["parent_uuid"]
                if org_unit_uuid in org_unit_map:
                    manager_uuid = org_unit_map[org_unit_uuid]["manager_uuid"]
                else:
                    # We are at the end of the upwards traversal
                    break
            result[employee_uuid] = manager_uuid

        return result

    def get_manager_uuid(self, mo_user: dict, eng_uuid) -> str | None:
        """Look up manager UUID for employee"""
        manager_uuid = self._manager_map.get(mo_user["uuid"])
        if manager_uuid == mo_user["uuid"]:
            return None
        return manager_uuid


class MORESTSource(MODataSource):
    """MO REST implementation of the MODataSource interface."""

    def __init__(self, settings):
        self.helper = MoraHelper(
            hostname=settings["global"]["mora.base"], use_cache=False
        )

    def read_user(self, uuid):
        mo_user = self.helper.read_user(user_uuid=uuid)
        if "uuid" not in mo_user:
            raise UserNotFoundException()
        else:
            assert mo_user["uuid"] == uuid
        exclude_fields = ["org", "user_key"]
        mo_user = dict_exclude(mo_user, exclude_fields)
        return mo_user

    def get_email_address(self, uuid):
        mail_dict = first(self.helper.get_e_addresses(uuid, scope="EMAIL"), default={})
        return dict_subset(mail_dict, ["uuid", "value"])

    def find_primary_engagement(self, uuid):
        def filter_primary(engagements):
            return filter(lambda eng: eng["is_primary"], engagements)

        user_engagements = self.helper.read_user_engagement(
            uuid, calculate_primary=True, read_all=True, skip_past=True
        )
        if not user_engagements:
            raise NoActiveEngagementsException()

        primary_engagement = next(filter_primary(user_engagements), None)
        if primary_engagement is None:
            raise NoPrimaryEngagementException("User: {}".format(uuid))

        employment_number = primary_engagement["user_key"]
        title = primary_engagement["job_function"]["name"]
        eng_org_unit = primary_engagement["org_unit"]["uuid"]
        eng_uuid = primary_engagement["uuid"]
        return employment_number, title, eng_org_unit, eng_uuid

    def get_manager_uuid(self, mo_user, eng_uuid):
        try:
            manager = self.helper.read_engagement_manager(eng_uuid)
            manager_uuid = manager["uuid"]
            return manager_uuid
        except KeyError:
            return None

    def get_engagement_dates(self, uuid):
        """Return all present and future engagement start dates and end dates.

        Args:
            uuid: UUID for the user to lookup.

        Returns:
            tuple[list[str], list[str]]: A tuple of lists of start and end dates.
        """
        user_engagements = self.helper.read_user_engagement(
            uuid, read_all=True, skip_past=True
        )
        dates = map(itemgetter("validity"), user_engagements)
        dates = map(itemgetter("from", "to"), dates)
        unzipped = unzip(dates)
        if len(unzipped) == 0:
            return [], []
        from_dates, to_dates = unzipped
        return from_dates, to_dates

    def get_it_systems(self, uuid):
        itsystems = self.helper.get_e_itsystems(uuid)

        def to_lora_itsystem(it_system):
            return it_system["itsystem"]["uuid"], {
                "uuid": it_system["uuid"],
                "user": it_system["person"]["uuid"],
                "unit": (it_system.get("org_unit") or {"uuid": None}).get("uuid"),
                "username": it_system["user_key"],
                "itsystem": it_system["itsystem"]["uuid"],
                "from_date": it_system["validity"]["from"],
                "to_date": it_system["validity"]["to"],
            }

        return dict(map(to_lora_itsystem, itsystems))


class ADWriter(AD):
    INVALID_UNIT_ADDRESS = {
        "city": INVALID,
        "postal_code": INVALID,
        "streetname": INVALID,
    }

    def __init__(self, lc=None, lc_historic=None, **kwargs):
        super().__init__(**kwargs)
        self.settings = self.all_settings
        self.skip_occupied_names = kwargs.get("skip_occupied_names", False)

        # Setup datasource for getting MO data.
        # TODO: Create a factory instead of this hackery?
        # Default to using MORESTSource as data source
        self.datasource = MORESTSource(self.settings)
        # Use LoraCacheSource if LoraCache is provided
        if lc and lc_historic:
            self.datasource = LoraCacheSource(lc, lc_historic, self.datasource)
        # NOTE: These should be eliminated when all uses are gone
        # NOTE: Once fully utilized, tests should be able to just implement a
        #       MODataSource for all their mocking needs.
        self.lc = lc
        self.lc_historic = lc_historic
        self.helper = MoraHelper(
            hostname=self.settings["global"]["mora.base"], use_cache=False
        )

        self._use_graphql_source_if_feature_flagged()
        self._init_name_creator()
        self._environment = self._get_jinja_environment()
        self._reader = ADParameterReader()

    def read_user(self, user=None, cpr=None):
        return self._reader.read_user(user=user, cpr=cpr)

    def _use_graphql_source_if_feature_flagged(self):
        feature_flag = self.settings.get("primary_write", {}).get("use_future_managers")
        if feature_flag:
            logger.info(
                "Using MOGraphqlSource to patch %r.get_manager_uuid", self.datasource
            )
            # Read manager information with graphql
            graphql = MOGraphqlSource(self.settings)
            # overwrite method no matter what datasource.
            self.datasource.get_manager_uuid = graphql.get_manager_uuid
        else:
            logger.info("Not using MOGraphqlSource")

    def _init_name_creator(self):
        self.name_creator = UserNameGen.get_implementation()
        if not self.skip_occupied_names:
            logger.info("Reading occupied names")
            self.name_creator.load_occupied_names()
        logger.info("Done reading occupied names")

    def _get_write_setting(self):
        if not self.all_settings["primary_write"]:
            msg = "Trying to enable write access with broken settings."
            logger.error(msg)
            raise Exception(msg)
        return self.all_settings["primary_write"]

    def _wait_for_replication(self, sam):
        # This method is only used by `ADWriter.create_user` (and only if called with
        # `create_manager=True`.) It is questionable whether `_wait_for_replication`
        # serves any real purpose, and we should consider removing it.

        t_start = time.time()
        logger.debug("Wait for replication of {}".format(sam))
        if not self.all_settings["global"]["servers"]:
            logger.info("No server infomation, falling back to waiting")
            time.sleep(15)
        else:
            # TODO, read from all AD servers and see when user is available
            replication_finished = False
            while not replication_finished:
                if time.time() - t_start > 60:
                    logger.error("Replication error")
                    raise ReplicationFailedException()

                for server in self.all_settings["global"]["servers"]:
                    user = self.read_user(user=sam, server=server)
                    logger.debug("Testing {}, found: {}".format(server, len(user)))
                    if user:
                        logger.debug("Found successfully")
                        replication_finished = True
                    else:
                        logger.debug("Did not find")
                        replication_finished = False
                        time.sleep(0.25)
                        break
        logger.info("replication_finished: {}s".format(time.time() - t_start))

    def _read_user(self, uuid):
        return self.datasource.read_user(uuid)

    def _find_unit_info(self, eng_org_unit):
        # TODO: Convert to datasource
        write_settings = self._get_write_setting()

        level2orgunit = "Ingen"
        unit_info = {}
        if self.lc:
            unit_name = self.lc.units[eng_org_unit][0]["name"]
            unit_user_key = self.lc.units[eng_org_unit][0]["user_key"]
            location = self.lc.units[eng_org_unit][0]["location"]

            # We initialize parent as the unit itself to ensure to catch if
            # a person is engaged directly in a level2org
            parent_uuid = self.lc.units[eng_org_unit][0]["uuid"]
            while parent_uuid is not None:
                parent_unit = self.lc.units[parent_uuid][0]
                if write_settings["level2orgunit_type"] in (
                    parent_unit["unit_type"],
                    parent_unit["level"],
                ):
                    level2orgunit = parent_unit["name"]
                parent_uuid = parent_unit["parent"]
        else:
            mo_unit_info = self.helper.read_ou(eng_org_unit)
            unit_name = mo_unit_info["name"]
            unit_user_key = mo_unit_info["user_key"]

            location = ""
            current_unit = mo_unit_info
            while current_unit:
                location = current_unit["name"] + "\\" + location
                current_type = current_unit["org_unit_type"]
                current_level = current_unit["org_unit_level"]
                if current_level is None:
                    current_level = {"uuid": None}
                if write_settings["level2orgunit_type"] in (
                    current_type["uuid"],
                    current_level["uuid"],
                ):
                    level2orgunit = current_unit["name"]
                current_unit = current_unit["parent"]
            location = location[:-1]

        unit_info = {
            "name": unit_name,
            "user_key": unit_user_key,
            "location": location,
            "level2orgunit": level2orgunit,
        }
        return unit_info

    def _read_user_addresses(self, eng_org_unit):
        # TODO: Convert to datasource
        addresses = {}
        if self.lc:
            email = []
            postal = {}
            for addr in self.lc.addresses.values():
                if addr[0]["unit"] == eng_org_unit:
                    if addr[0]["scope"] == "DAR":
                        postal = {"Adresse": addr[0]["value"]}
                    if addr[0]["scope"] == "E-mail":
                        visibility = addr[0]["visibility"]
                        visibility_class = None
                        if visibility is not None:
                            visibility_class = self.lc.classes[visibility]
                        email.append(
                            {"visibility": visibility_class, "value": addr[0]["value"]}
                        )
        else:
            email = self.helper.read_ou_address(
                eng_org_unit, scope="EMAIL", return_all=True
            )
            postal = self.helper.read_ou_address(
                eng_org_unit, scope="DAR", return_all=False
            )

        unit_secure_email = None
        unit_public_email = None
        for mail in email:
            if mail["visibility"] is None:
                # If visibility is not set, we assume it is non-public.
                unit_secure_email = mail["value"]
            else:
                if mail["visibility"]["scope"] == "PUBLIC":
                    unit_public_email = mail["value"]
                if mail["visibility"]["scope"] == "SECRET":
                    unit_secure_email = mail["value"]

        addresses = {
            "unit_secure_email": unit_secure_email,
            "unit_public_email": unit_public_email,
            "postal": postal,
        }
        return addresses

    def read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        """
        Retrive the necessary information from MO to contruct a new AD user.
        The final information object should of this type, notice that end-date
        is not necessarily for the current primary engagement, but the end-date
        of the longest running currently known primary engagement:
        mo_values = {
            'name': ('Martin Lee', 'Gore'),
            'employment_number': '101',
            'uuid': '7ccbd9aa-gd60-4fa1-4571-0e6f41f6ebc0',
            'end_date': 2089-11-11,
            'cpr': '1122334455',
            'title': 'Musiker',
            'location': 'Viborg Kommune\\Forvalting\\Enhed\\',
            'level2orgunit: 'Beskæftigelse, Økonomi & Personale',
            'manager_sam': 'DMILL'
        }
        """
        logger.info("Read information for {}".format(uuid))
        try:
            (
                employment_number,
                title,
                eng_org_unit,
                eng_uuid,
            ) = self.datasource.find_primary_engagement(uuid)
        except NoActiveEngagementsException:
            logger.info("No active engagements found")
            return None

        unit_info = self._find_unit_info(eng_org_unit)
        if self._skip_unit(unit_info):
            logger.info("Skipping %r (location=%r)", uuid, unit_info["location"])
            return None

        def split_addresses(addresses):
            try:
                postal = addresses["postal"]["Adresse"]
            except KeyError:
                return ADWriter.INVALID_UNIT_ADDRESS
            else:
                try:
                    postal_code = re.findall("[0-9]{4}", postal)[0]
                    city_pos = postal.find(postal_code) + 5
                    city = postal[city_pos:]
                    streetname = postal[: city_pos - 7]
                except (IndexError, TypeError):
                    logger.error("Unable to read adresse from MO (no access to DAR?)")
                    return ADWriter.INVALID_UNIT_ADDRESS
                else:
                    return {
                        "postal_code": postal_code,
                        "city": city,
                        "streetname": streetname,
                    }

        def read_manager_uuid(mo_user, eng_uuid):
            manager_uuid = self.datasource.get_manager_uuid(mo_user, eng_uuid)
            if manager_uuid is None:
                logger.info("No managers found")
            return manager_uuid

        def read_manager_mail(manager_uuid):
            manager_mail_dict = self.datasource.get_email_address(manager_uuid)
            if manager_mail_dict:
                return manager_mail_dict["value"]
            return None

        def read_manager_sam(manager_cpr):
            try:
                manager_ad_user = self._find_ad_user(cpr=manager_cpr, ad_dump=ad_dump)
            except CprNotFoundInADException:
                logger.info("manager not found by cpr lookup")
            except CprNotNotUnique:
                logger.info("multiple managers found by cpr lookup")
            else:
                return self._get_sam_from_ad_values(manager_ad_user)
            return None

        # NOTE: Underscore fields should not be read
        mo_values: LazyDict = LazyDict(
            {
                # Raw information
                "uuid": uuid,
                # Engagement information
                "employment_number": employment_number,
                "title": title,
                "unit_uuid": eng_org_unit,
                "_eng_uuid": eng_uuid,
                "_dates": LazyEvalDerived(
                    lambda uuid: self.datasource.get_engagement_endpoint_dates(uuid)
                ),
                "start_date": LazyEvalDerived(lambda _dates: _dates[0]),
                "end_date": LazyEvalDerived(lambda _dates: _dates[1]),
                # Lazy MO User and associated fields
                "_mo_user": LazyEvalDerived(lambda uuid: self._read_user(uuid)),
                "name": LazyEvalDerived(
                    lambda _mo_user: (_mo_user["givenname"], _mo_user["surname"])
                ),
                "full_name": LazyEvalDerived(
                    lambda name: "{} {}".format(*name).strip()
                ),
                "nickname": LazyEvalDerived(
                    lambda _mo_user: (
                        _mo_user["nickname_givenname"],
                        _mo_user["nickname_surname"],
                    )
                ),
                "full_nickname": LazyEvalDerived(
                    lambda nickname: "{} {}".format(*nickname).strip()
                ),
                "cpr": LazyEvalDerived(lambda _mo_user: _mo_user["cpr_no"]),
                # Lazy Unit and associated fields
                "_unit": LazyEvalDerived(
                    lambda unit_uuid: self._find_unit_info(unit_uuid)
                ),
                "unit": LazyEvalDerived(lambda _unit: _unit["name"]),
                "unit_user_key": LazyEvalDerived(lambda _unit: _unit["user_key"]),
                "location": LazyEvalDerived(lambda _unit: _unit["location"]),
                "level2orgunit": LazyEvalDerived(lambda _unit: _unit["level2orgunit"]),
                # Lazy addresses and associated fields
                "_addresses": LazyEvalDerived(
                    lambda unit_uuid: self._read_user_addresses(unit_uuid)
                ),
                "_parsed_addresses": LazyEvalDerived(
                    lambda _addresses: split_addresses(_addresses)
                ),
                "unit_postal_code": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["postal_code"]
                ),
                "unit_city": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["city"]
                ),
                "unit_streetname": LazyEvalDerived(
                    lambda _parsed_addresses: _parsed_addresses["streetname"]
                ),
                "unit_public_email": LazyEvalDerived(
                    lambda _addresses: _addresses["unit_public_email"]
                ),
                "unit_secure_email": LazyEvalDerived(
                    lambda _addresses: _addresses["unit_secure_email"]
                ),
                # Manager stuff
                "_manager_uuid": LazyEval(
                    lambda key, dictionary: (
                        read_manager_uuid(
                            dictionary["_mo_user"], dictionary["_eng_uuid"]
                        )
                        if read_manager
                        else None
                    )
                ),
                "_manager_mo_user": LazyEvalDerived(
                    lambda _manager_uuid: self._read_user(_manager_uuid)
                    if _manager_uuid
                    else {}
                ),
                "manager_name": LazyEvalDerived(
                    lambda _manager_mo_user: _manager_mo_user.get("name")
                ),
                "manager_cpr": LazyEvalDerived(
                    lambda _manager_mo_user: _manager_mo_user.get("cpr_no")
                ),
                "manager_mail": LazyEvalDerived(
                    lambda _manager_uuid: read_manager_mail(_manager_uuid)
                    if _manager_uuid
                    else None
                ),
                "manager_sam": LazyEvalDerived(
                    lambda manager_cpr: read_manager_sam(manager_cpr)
                    if manager_cpr
                    else None
                ),
                "read_manager": LazyEvalDerived(
                    lambda _manager_uuid: bool(_manager_uuid)
                ),
                # Employee addresses
                "employee_addresses": LazyEvalDerived(
                    lambda uuid: self.helper.get_e_addresses(uuid)
                ),
                # IT systems
                "it_systems": LazyEvalDerived(
                    lambda uuid: self.datasource.get_it_systems(uuid)
                ),
            }
        )
        return mo_values

    def add_manager_to_user(self, user_sam, manager_sam):
        """
        Mark an existing AD user as manager for an existing AD user.
        :param user_sam: SamAccountName for the employee.
        :param manager_sam: SamAccountName for the manager.
        """
        ps_script = self._get_add_manager_command(user_sam, manager_sam)
        response = self._run_ps_script(ps_script)
        return response is {}

    def _get_add_manager_command(self, user_sam, manager_sam):
        format_rules = {"user_sam": user_sam, "manager_sam": manager_sam}
        ps_script = self._build_ps(ad_templates.add_manager_template, format_rules)
        return ps_script

    def _rename_ad_user(self, user_sam, new_name):
        logger.info("Rename user: %s", user_sam)
        ps_script = self._get_rename_ad_user_command(user_sam, new_name)
        logger.debug("Rename user, ps_script: {}".format(ps_script))
        response = self._run_ps_script(ps_script)
        logger.debug("Response from sync: {}".format(response))
        logger.debug("Wait for replication")
        # Todo: In principle we should ask all DCs, bu this will happen
        # very rarely, performance is not of great importance
        time.sleep(1)

    def _get_rename_ad_user_command(self, user_sam, new_name):
        # Todo: This code is a duplicate of code found elsewhere
        # Todo: This code is buggy - it uses a "stringified tuple" as the new
        # AD username. E.g. the AD user is renamed to
        # `"(\"Firstname\", \"Lastname\")"`.
        rename_user_template = ad_templates.rename_user_template
        rename_user_string = rename_user_template.format(
            user_sam=user_sam,
            new_name=new_name,
        )
        rename_user_string = self.remove_redundant(rename_user_string)
        server_string = ""
        if self.all_settings["global"].get("servers") is not None:
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )
        ps_script = self._build_user_credential() + rename_user_string + server_string
        return ps_script

    def _compare_fields(self, ad_field, value, ad_user):
        mismatch = {}
        ad_field_value = ad_user.get(ad_field)

        # Some AD fields contain one-element lists, rather than the usual strings,
        # numbers or UUIDs.
        # In such cases, we "unpack" the single-element list before comparing it to the
        # corresponding MO value - otherwise the comparison will not work as expected.
        if isinstance(ad_field_value, list) and len(ad_field_value) == 1:
            ad_field_value = ad_field_value[0]

        # The "MO value" in `value` actually comes from a call to
        # `_render_field_template`, which may produce the string `"None"` when it
        # encounters a `None`.
        if value == "None":
            value = None

        # We also consider the reverse situation where the AD field contains the string
        # "None" to indicate an empty value.
        if ad_field_value == "None":
            ad_field_value = None

        # Do the actual comparison
        if value == INVALID:
            msg = "%r: MO value is INVALID, not changing AD value %r"
            logger.info(msg, ad_field, ad_field_value)
        elif ad_field_value != value:
            msg = "%r: AD value %r does not match MO value %r"
            logger.info(msg, ad_field, ad_field_value, value)
            mismatch = {ad_field: (ad_field_value, value)}
        else:
            msg = "%r: AD value %r already matches MO value %r"
            logger.debug(msg, ad_field, ad_field_value, value)

        return mismatch

    def _get_jinja_environment(self):
        environment = Environment(undefined=StrictUndefined)

        # Add custom filters
        environment.filters["first_address_of_type"] = first_address_of_type
        environment.filters["name_to_email_address"] = name_to_email_address
        environment.filters["location_element"] = location_element

        @lru_cache
        def get_all_ad_emails():
            reader = ADParameterReader(all_settings=self.all_settings)
            return reader.get_all_email_values()

        # Add globally available vars
        environment.globals["_upn_end"] = self.settings["primary_write"]["upn_end"]
        environment.globals["_get_all_ad_emails"] = get_all_ad_emails

        return environment

    def _render_field_template(self, context, template):
        env = self._environment.overlay(undefined=Undefined)
        template = env.from_string(template.strip('"'))
        return template.render(**context)

    def _preview_create_command(self, mo_uuid, ad_dump=None, create_manager=True):
        mo_values = self.read_ad_information_from_mo(
            mo_uuid,
            ad_dump=ad_dump,
            read_manager=create_manager,
        )
        sam_account_name = self._get_create_user_sam_account_name(
            mo_values, dry_run=True
        )
        create_cmd = self._get_create_user_command(mo_values, sam_account_name)
        add_manager_cmd = self._get_add_manager_command(
            sam_account_name, mo_values["manager_sam"]
        )
        return create_cmd, add_manager_cmd

    def _preview_sync_command(self, mo_uuid, user_sam, ad_dump=None, sync_manager=True):
        mo_values = self.read_ad_information_from_mo(
            mo_uuid, ad_dump=ad_dump, read_manager=sync_manager
        )

        try:
            ad_values = self._find_ad_user(mo_values["cpr"], ad_dump=ad_dump)
        except CprNotFoundInADException:
            ad_values = {}
        else:
            user_sam = self._get_sam_from_ad_values(ad_values)

        sync_cmd = self._get_sync_user_command(ad_values, mo_values, user_sam)
        rename_cmd = ""
        rename_cmd_target = ""
        add_manager_cmd = ""

        try:
            mismatch = self._sync_compare(mo_values, ad_dump)
        except CprNotFoundInADException:
            # We might be running against an AD where we have not yet written any actual
            # users. In that case, we will not find the user by CPR. But we still want
            # to see the rename command that would be issued.
            logger.info("Previewing 'rename' command for nonexistent AD user")
            rename_cmd = self._get_rename_ad_user_command(user_sam, "<new name>")
            rename_cmd_target = "<nonexistent AD user>"
        else:
            # We found an actual AD user by CPR

            # Preview 'rename' command, if the name differs between MO and AD
            if "name" in mismatch:
                # A rename command is necessary, as the new username differs from the
                # current username in AD.
                logger.info("Previewing 'rename' command for existing AD user")
                new_name = mismatch["name"][1]
                rename_cmd = self._get_rename_ad_user_command(user_sam, new_name)
                rename_cmd_target = mismatch["name"][0]  # = previous username

            # Preview 'add manager' command, if the manager differs between MO and AD
            if sync_manager and "manager" in mismatch:
                add_manager_cmd = self._get_add_manager_command(
                    user_sam, mo_values["manager_sam"]
                )

        return sync_cmd, rename_cmd, rename_cmd_target, add_manager_cmd

    def _sync_compare(self, mo_values, ad_dump):
        ad_user = self._find_ad_user(mo_values["cpr"], ad_dump=ad_dump)
        user_sam = self._get_sam_from_ad_values(ad_user)

        # TODO: Why is this not generated along with all other info in mo_values?
        mo_values["name_sam"] = "{} - {}".format(mo_values["full_name"], user_sam)

        fields = prepare_field_templates("Set-ADUser", settings=self.all_settings)

        def to_lower(string):
            return string.lower()

        ad_user = dict_map(ad_user, key_func=to_lower)
        fields = dict_map(fields, key_func=to_lower)

        never_compare = lower_list(["Credential", "Manager"])
        fields = dict_exclude(fields, never_compare)

        context = {
            "ad_values": ad_user,
            "mo_values": mo_values,
            "user_sam": user_sam,
        }

        # Build context and render template to get comparision value
        # NOTE: This results in rendering the template twice, once here and
        #       once inside the powershell render call.
        #       We should probably restructure this, such that we only render
        #       the template once, potentially rendering a dict of results.
        # TODO: Make the above mentioned change.
        fields = dict_map(
            fields,
            value_func=partial(self._render_field_template, context),
        )
        mismatch = {}
        for ad_field, rendered_value in fields.items():
            mismatch.update(self._compare_fields(ad_field, rendered_value, ad_user))

        if mo_values.get("manager_cpr"):
            try:
                manager_ad_user = self._find_ad_user(
                    mo_values["manager_cpr"], ad_dump=ad_dump
                )
            except CprNotFoundInADException:
                logger.warning("could not find manager by cpr")
            else:
                manager_distinguished_name = manager_ad_user["DistinguishedName"]
                if ad_user["manager"] != manager_distinguished_name:
                    mismatch["manager"] = (
                        ad_user["manager"],
                        manager_distinguished_name,
                    )
                    logger.info("Manager should be updated")
                else:
                    logger.debug("Not updating manager")

        return mismatch

    def sync_user(self, mo_uuid, ad_dump=None, sync_manager=True):
        """
        Sync MO information into AD
        """
        mo_values = self.read_ad_information_from_mo(
            mo_uuid, ad_dump=ad_dump, read_manager=sync_manager
        )

        if mo_values is None:
            return (False, "No active engagments")

        ad_values = self._find_ad_user(mo_values["cpr"], ad_dump=ad_dump)
        user_sam = self._get_sam_from_ad_values(ad_values)

        if ad_dump is None:
            # TODO: We could also add the compare logic here,
            # but the benefit will be max 40%
            mismatch = {"force re-sync": "yes", "manager": "yes"}
        else:
            mismatch = self._sync_compare(mo_values, ad_dump)

        logger.debug("Sync compare: {}".format(mismatch))

        if "name" in mismatch:
            response = self._rename_ad_user(user_sam, mismatch["name"][1])
            del mismatch["name"]

        if not mismatch and ("sync_timestamp" not in str(self.all_settings)):
            # If "sync_timestamp" is in settings we assume the intent is to always write a timestamp.
            logger.info("Nothing to edit")
            return (True, "Nothing to edit", mo_values["read_manager"])

        logger.info("Sync compare: {}".format(mismatch))

        ps_script = self._get_sync_user_command(ad_values, mo_values, user_sam)
        logger.debug("Sync user, ps_script: {}".format(ps_script))

        response = self._run_ps_script(ps_script)
        logger.debug("Response from sync: {}".format(response))

        if sync_manager and "manager" in mismatch:
            logger.info("Add manager")
            self.add_manager_to_user(
                user_sam=user_sam, manager_sam=mo_values["manager_sam"]
            )

        return (True, "Sync completed", mo_values["read_manager"])

    def _get_sync_user_command(self, ad_values, mo_values, user_sam):
        edit_user_string = template_powershell(
            cmd="Set-ADUser",
            context={
                "ad_values": ad_values,
                "mo_values": mo_values,
                "user_sam": user_sam,
                "sync_timestamp": str(datetime.now()),
            },
            settings=self.all_settings,
            environment=self._environment,
        )
        edit_user_string = self.remove_redundant(edit_user_string)

        server_string = ""
        if self.all_settings["global"].get("servers") is not None:
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )

        ps_script = self._build_user_credential() + edit_user_string + server_string
        return ps_script

    def create_user(self, mo_uuid, create_manager, dry_run=False):
        """
        Create an AD user
        :param mo_uuid: uuid for the MO user we want to add to AD.
        :param create_manager: If True, an AD link will be added between the user
        object and the AD object of the users manager.
        :param dry_run: generates a username and checks wheter the user exists in AD.
        :return: The generated SamAccountName for the new user
        """
        mo_values = self.read_ad_information_from_mo(mo_uuid, create_manager)
        if mo_values is None:
            logger.error("Trying to create user with no engagements")
            raise NoPrimaryEngagementException

        sam_account_name = self._get_create_user_sam_account_name(
            mo_values, dry_run=dry_run
        )

        self._check_if_ad_user_exists(sam_account_name, mo_values["cpr"])

        ps_script = self._get_create_user_command(mo_values, sam_account_name)
        if dry_run:
            return (True, sam_account_name)

        response = self._run_ps_script(ps_script)
        if not response == {}:
            msg = "Create user failed, message: {}".format(response)
            logger.error(msg)
            return (False, msg)

        if create_manager:
            self._wait_for_replication(sam_account_name)
            msg = "Add {} as manager for {}".format(
                mo_values["manager_sam"], sam_account_name
            )
            print(msg)
            logger.info(msg)
            self.add_manager_to_user(
                user_sam=sam_account_name, manager_sam=mo_values["manager_sam"]
            )

        return (True, sam_account_name)

    def _get_create_user_sam_account_name(self, mo_values, dry_run=False):
        all_names = mo_values["name"][0].split(" ") + [mo_values["name"][1]]
        return self.name_creator.create_username(all_names, dry_run=dry_run)

    def _get_create_user_command(self, mo_values, sam_account_name):
        create_user_string = template_powershell(
            context={
                "ad_values": {},
                "mo_values": mo_values,
                "user_sam": sam_account_name,
                "sync_timestamp": str(datetime.now()),
            },
            settings=self.all_settings,
            environment=self._environment,
        )
        create_user_string = self.remove_redundant(create_user_string)

        # Should this go to self._ps_boiler_plate()?
        server_string = ""
        if self.all_settings["global"].get("servers"):
            server_string = " -Server {} ".format(
                random.choice(self.all_settings["global"]["servers"])
            )

        ps_script = (
            self._build_user_credential()
            + create_user_string
            + server_string
            + self._get_new_ad_user_path_argument()
        )

        return ps_script

    def _get_new_ad_user_path_argument(self):
        primary = self._get_setting()
        primary_write = self._get_write_setting()
        path = primary_write.get("new_ad_user_path", primary["search_base"])
        path_argument = f' -Path "{path}"'
        return path_argument

    def _check_if_ad_user_exists(self, sam_account_name, cpr):
        if sam_account_name and self.read_user(user=sam_account_name):
            logger.error("SamAccount already in use: {}".format(sam_account_name))
            raise SamAccountNameNotUnique(sam_account_name)
        if cpr and self.read_user(cpr=cpr):
            logger.error(f"cpr already in use: {cpr[:6]}-xxxx")
            raise CprNotNotUnique()

    def _get_enable_user_cmd(self, username: str, enable: bool) -> str:
        # Hack: copy the settings and mutate `template_to_ad_fields` in the copy.
        # This enables us to use a different set of field templates when enabling and
        # disabling AD users.
        settings = copy.deepcopy(self.settings)
        template_to_ad_fields = {"Enabled": "$true" if enable else "$false"}
        if enable is False:
            # Read an additional field mapping which is only to be used when disabling
            # AD users.
            template_to_ad_fields_when_disable = settings["primary_write"].get(
                "template_to_ad_fields_when_disable", {}
            )
            template_to_ad_fields.update(template_to_ad_fields_when_disable)
        settings["primary_write"]["template_to_ad_fields"] = template_to_ad_fields

        # Render the PowerShell command to enable or disable an AD user
        cmd = template_powershell(
            cmd="Set-ADUser",
            context={
                "user_sam": username,
                # Available for the template rendering.
                "now": datetime.now(),
                # Required by `template_powershell`, but not actually used.
                "mo_values": {"level2orgunit": None, "location": None},
            },
            settings=settings,
        )
        cmd = self._build_user_credential() + self.remove_redundant(cmd)
        return cmd

    def enable_user(self, username, enable=True):
        """
        Enable or disable an AD account.
        :param username: SamAccountName of the account to be enabled or disabled
        :param enable: If True enable account, if False, disable account
        """
        cmd = self._get_enable_user_cmd(username, enable)
        response = self._run_ps_script(cmd)  # raises `CommandFailure` if PS fails
        if response == {}:
            return True, "enabled AD user" if enable else "disabled AD user"

    def _skip_unit(self, unit_info):
        skip_locations = self.all_settings.get("primary_write", {}).get(
            "skip_locations"
        )
        if skip_locations is None:
            return False  # no skip locations defined
        if unit_info.get("location"):
            skip_locations_lower = set(s.lower() for s in skip_locations)
            location = unit_info["location"]
            parts = location.split("\\")
            if any(part.lower() in skip_locations_lower for part in parts):
                return True


@click.command()
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option(
    "--create-user-with-manager",
    help="Create a new user in AD, also assign a manager",
)
@optgroup.option(
    "--create-user",
    help="Create a new user in AD, do not assign a manager",
)
@optgroup.option(
    "--sync-user",
    help="Sync relevant fields from MO to AD",
)
@optgroup.option(
    "--mo-values",
    type=click.UUID,
    help="Show mo-values for the user",
)
@optgroup.option("--read-ad-information")
@optgroup.option("--add-manager-to-user", nargs=2, type=str)
@click.option("--ignore-occupied-names", is_flag=True, default=False)
def cli(**args):
    """
    Command line interface for the AD writer class.
    """

    ad_writer = ADWriter(skip_occupied_names=args["ignore_occupied_names"])

    if args.get("create_user_with_manager"):
        print("Create_user_with_manager:")
        status = ad_writer.create_user(
            args["create_user_with_manager"], create_manager=True
        )
        # TODO: execute custom script? Or should this be done in
        # two steps.
        print(status[1])

    if args.get("create_user"):
        print("Create user, no link to manager:")
        status = ad_writer.create_user(args["create_user"], create_manager=False)
        print(status[1])

    if args.get("sync_user"):
        print("Sync MO fields to AD")
        status = ad_writer.sync_user(args["sync_user"])
        print(status[1])

    if args.get("read_ad_information"):
        print("AD information on user:")
        sam = args["read_ad_information"]
        user = ad_writer.read_user(user=sam)
        if not user:
            print("User not found")
        else:
            for key, value in sorted(user[0].items()):
                print("{}: {}".format(key, value))

    if args.get("add_manager_to_user"):
        manager, user = args["add_manager_to_user"]
        print("{} is now set as manager for {}".format(manager, user))
        ad_writer.add_manager_to_user(manager_sam=manager, user_sam=user)

    if args.get("mo_values"):
        mo_values = ad_writer.read_ad_information_from_mo(str(args["mo_values"]))
        print(json.dumps(dict(mo_values.items()), indent=4))


if __name__ == "__main__":
    start_logging()
    cli()



import time
import uuid
from datetime import datetime
from functools import partial
from itertools import chain
from random import choice
from random import randint
from unittest.mock import patch

from ..ad_sync import AdMoSync
from ..ad_writer import ADWriter
from ..read_ad_conf_settings import read_settings
from ..user_names import UserNameGen
from ..utils import AttrDict
from ..utils import recursive_dict_update
from .name_simulator import create_name


def dict_modifier(updates):
    """Wrapper around recursive_dict_update, which provides updates beforehand.

    Example:

        updates = {'action': 'set'}

        function_with_expects_single_argument_transformer(
            dict_modifier(updates)
        )

    Args:
        updates: dictionary with updates to be applied later, when the returned
                 function is actually called.

    Returns:
        function: A partially applied recursive_dict_update function, waiting
                  for the original to apply updates to.
    """
    return partial(recursive_dict_update, updates=updates)


def mo_modifier(updates):
    """Wrapper around recursive_dict_update, which provides updates beforehand.

    Example:

        updates = {'action': 'set'}

        function_with_expects_mo_values_call(
            mo_modifier(updates)
        )

    Args:
        updates: dictionary with updates to be applied later, when the returned
                 function is actually called.

    Returns:
        function: A partially applied recursive_dict_update function, waiting
                  for the original to apply updates to.
    """

    def mo_mod(mo_values, *args, **kwargs):
        return recursive_dict_update(mo_values, updates=updates)

    return mo_mod


class ADWriterTestSubclass(ADWriter):
    """Testing subclass of ADWriter."""

    def __init__(
        self,
        read_ad_information_from_mo,
        ad_values_func=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # List of scripts to be executed via run_ps
        self.scripts = []
        # Transformer for mo_values return
        self.read_ad_information_from_mo = read_ad_information_from_mo
        # Replace real `_find_ad_user` with mock
        if kwargs.get("mock_find_ad_user", True):
            self._find_ad_user = lambda ad_user, ad_dump=None: ad_values_func()

    def _init_name_creator(self):
        """Mocked to pretend no names are occupied.

        This method would normally use ADReader to read usernames from AD.
        """
        self.name_creator = UserNameGen.get_implementation()

    def _create_session(self):
        """Mocked to return a fake-class which writes scripts to self.scripts.

        This method would normally send scripts to powershell via WinRM.
        """

        def run_ps(ps_script):
            # Add our script to the list
            self.scripts.append(ps_script)
            # Fake the WinRM run_ps return type
            return AttrDict(
                {
                    "status_code": 0,
                    "std_out": b"",
                    "std_err": b"",
                }
            )

        # Fake the WinRM session object
        return AttrDict(
            {
                "run_ps": run_ps,
            }
        )

    def _get_retry_exceptions(self):
        """Mocked to return an empty list, i.e. never retry.

        This method would normally return the WinRM transport exception, to
        cause retrying to happen.
        """
        return []

    def read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        raise NotImplementedError("Should be overridden in __init__")

    def _read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        return super().read_ad_information_from_mo(uuid, read_manager, ad_dump)


def _no_transformation(default, *args, **kwargs):
    return default


class TestADMixin(object):
    # Useful for local testing
    generate_dynamic_person = True
    default_person = None

    def _prepare_dynamic_person(self):
        def random_date():
            unixtime = randint(1, int(time.time()))
            return datetime.fromtimestamp(unixtime).strftime("%d%m%y")

        def random_digit():
            return choice("0123456789")

        def random_digits(num_digits):
            return "".join(random_digit() for _ in range(num_digits))

        def gen_cpr():
            return random_date() + random_digits(4)

        dynamic_person = {
            "uuid": str(uuid.uuid4()),
            "name": create_name(),
            "nickname": create_name(),
            "cpr": gen_cpr(),
            "employment_number": random_digits(4),
            "read_manager": True,
            "manager_name": create_name(),
            "manager_cpr": gen_cpr(),
        }
        return dynamic_person

    def _prepare_static_person(self):
        static_person = {
            "uuid": "7ccbd9aa-gd60-4fa1-4571-0e6f41f6ebc0",
            "name": ("Martin", "Lee", "Gore"),
            "nickname": ("Depeche", "Mode"),
            "employment_number": "101",
            "cpr": "1122334455",
            "read_manager": True,
            "manager_name": ("Daniel", "Miller"),
            "manager_cpr": "1122334455",
        }
        return static_person

    def _prepare_person(self, person_transformer=None, *args, **kwargs):
        if self.default_person is None:
            default_person = self._prepare_static_person()
            if self.generate_dynamic_person:
                default_person = self._prepare_dynamic_person()
            # Add computed fields
            creator = UserNameGen.get_implementation()
            sam_account_name = creator.create_username(list(default_person["name"]))
            default_person.update(
                **{
                    "full_name": " ".join(default_person["name"]),
                    "sam_account_name": sam_account_name,
                    "manager_sam": default_person["manager_name"][0],
                    "manager_email": default_person["manager_name"][0] + "@magenta.dk",
                }
            )
            # Add static fields
            default_person.update(
                **{
                    "end_date": "2089-11-11",
                    "title": "Musiker",
                    "unit": "Enhed",
                    "unit_uuid": "101bd9aa-0101-0101-0101-0e6f41f6ebc0",
                    "unit_user_key": "Musik",
                    "unit_postal_code": "8210",
                    "unit_city": "Aarhus N",
                    "unit_streetname": "Fahrenheit 451",
                    "location": "Kommune\\Forvalting\\Enhed\\",
                    "level2orgunit": "Ingen",
                    "forvaltning": "Beskæftigelse, Økonomi & Personale",
                }
            )
            transformer_func = person_transformer or _no_transformation
            self.default_person = transformer_func(default_person, *args, **kwargs)
        return self.default_person

    def _prepare_mo_values(self, mo_values_transformer=None, *args, **kwargs):
        person = self._prepare_person()
        # Convert raw person data into mo_values data
        person["name"] = [" ".join(person["name"][:-1]), person["name"][-1]]
        person["nickname"] = [
            " ".join(person["nickname"][:-1]),
            person["nickname"][-1],
        ]
        person["manager_name"] = " ".join(person["manager_name"])

        #        if not read_manager:
        #            del person['manager_name']
        #            del person['manager_sam']
        #            del person['manager_email']
        #            del person['manager_cpr']
        #           person["read_manager"] = False

        transformer_func = mo_values_transformer or _no_transformation
        return transformer_func(person, *args, **kwargs)

    def _prepare_get_from_ad(self, ad_transformer, *args, **kwargs):
        person = self._prepare_person()
        # Convert raw person data into ad_values data
        default_ad_person = {
            "ObjectGUID": person["uuid"],
            "SID": {
                "AccountDomainSid": {
                    "AccountDomainSid": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx",
                    "BinaryLength": 24,
                    "Value": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx",
                },
                "BinaryLength": 28,
                "Value": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx-xxxxx",
            },
            "PropertyCount": 11,
            "PropertyNames": [
                "ObjectGUID",
                "SID",
                "DistinguishedName",
                "Enabled",
                "GivenName",
                "Name",
                "ObjectClass",
                "SamAccountName",
                "Surname",
                "UserPrincipalName",
                "extensionAttribute1",
            ],
            "DistinguishedName": "CN="
            + person["full_name"]
            + ",OU="
            + person["unit"]
            + ",DC=lee",
            "Enabled": True,
            "Name": person["full_name"],
            "ObjectClass": "user",
            "SamAccountName": person["sam_account_name"],
            "GivenName": person["name"][-1],
            "UserPrincipalName": "_".join(person["full_name"]).lower() + "@magenta.dk",
            "extensionAttribute1": person["cpr"],
            "AddedProperties": [],
            "ModifiedProperties": [],
            "RemovedProperties": [],
        }
        transformer_func = ad_transformer or _no_transformation
        return transformer_func(default_ad_person, *args, **kwargs)

    def _prepare_settings(self, early_settings_transformer=None):
        """Load default settings for AD tests.

        Args:
            early_settings_transformer: Function to transform settings.

        Returns:
            dict: Default settings after transformation.
        """
        default_settings = {
            "integrations.ad": [
                {
                    "cpr_field": "cpr_field",
                    "cpr_separator": "ad_cpr_sep",
                    "system_user": "system_user",
                    "password": "password",
                    "properties": [],
                    "search_base": "search_base",
                    "integrations.ad.ad_mo_sync_mapping": {},
                    "ad_mo_sync_terminate_missing": False,
                    "ad_mo_sync_terminate_missing_require_itsystem": True,
                    "ad_mo_sync_terminate_disabled": True,
                    "ad_mo_sync_pre_filters": [],
                    "ad_mo_sync_terminate_disabled_filters": [],
                    "servers": ["server123"],
                }
            ],
            "integrations.ad.winrm_host": "dummy",
            # "integrations.ad.sam_filter": "sam_filter",
            "mora.base": "http://example.org",
            "integrations.ad.write.uuid_field": "uuid_field",
            "integrations.ad.write.level2orgunit_field": "level2orgunit_field",
            "integrations.ad.write.org_unit_field": "org_field",
            "integrations.ad.write.upn_end": "epn_end",
            "integrations.ad.write.level2orgunit_type": "level2orgunit_type",
            "integrations.ad_writer.template_to_ad_fields": {
                "Name": "{{ mo_values['full_name'] }} - {{ user_sam }}",
                "Displayname": "{{ mo_values['name'][0] }} {{ mo_values['name'][1] }}",
                "GivenName": "{{ mo_values['name'][0] }}",
                "SurName": "{{ mo_values['name'][1] }}",
                "EmployeeNumber": "{{ mo_values['employment_number'] }}",
            },
            "address.visibility.public": "address_visibility_public_uuid",
            "address.visibility.internal": "address_visibility_internal_uuid",
            "address.visibility.secret": "address_visibility_secret_uuid",
        }
        transformer_func = early_settings_transformer or _no_transformation
        modified_settings = transformer_func(default_settings)
        for ad_settings in modified_settings["integrations.ad"]:
            ad_settings["properties"] = list(
                map(
                    lambda x: x.lower(),
                    chain(
                        modified_settings.get(
                            "integrations.ad_writer.template_to_ad_fields", {}
                        ).keys(),
                        modified_settings.get(
                            "integrations.ad_writer.mo_to_ad_fields", {}
                        ).values(),
                        [modified_settings["integrations.ad.write.org_unit_field"]],
                        [
                            modified_settings[
                                "integrations.ad.write.level2orgunit_field"
                            ]
                        ],
                        [modified_settings["integrations.ad.write.uuid_field"]],
                    ),
                )
            )
        return modified_settings


class TestADWriterMixin(TestADMixin):
    def _setup_adwriter(
        self,
        late_transform_settings=None,
        transform_mo_values=None,
        early_transform_settings=None,
        transform_ad_values=lambda x: x,
        **kwargs,
    ):
        transformer_func = late_transform_settings or _no_transformation
        self.settings = transformer_func(
            read_settings(self._prepare_settings(early_transform_settings))
        )
        self.mo_values_func = partial(self._prepare_mo_values, transform_mo_values)
        self.ad_values_func = partial(self._prepare_get_from_ad, transform_ad_values)

        # Avoid circular imports
        from .mocks import MockEmptyADReader
        from .mocks import MockMOGraphqlSource

        with patch(
            "integrations.ad_integration.ad_writer.MOGraphqlSource",
            new=MockMOGraphqlSource,
        ):
            with patch(
                "integrations.ad_integration.ad_writer.ADParameterReader",
                kwargs.get("mock_ad_reader_class", MockEmptyADReader),
            ):
                self.ad_writer = ADWriterTestSubclass(
                    all_settings=self.settings,
                    read_ad_information_from_mo=self.mo_values_func,
                    ad_values_func=self.ad_values_func,
                    **kwargs,
                )


class AdMoSyncTestSubclass(AdMoSync):
    def __init__(
        self,
        mo_values_func,
        mo_e_username_func,
        ad_values_func,
        mo_seed_func,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.mo_values = mo_values_func()
        self.ad_values = ad_values_func()
        self.e_username = mo_e_username_func()

        self.mo_seed = mo_seed_func()
        self.mo_post_calls = []

    def _verify_it_systems(self):
        pass

    def _setup_mora_helper(self):
        def get_e_details(detail_type):
            return lambda *args, **kwargs: self.mo_seed[detail_type]

        def _mo_post(url, payload, force=True):
            # Register the call, so we can test against it
            self.mo_post_calls.append({"url": url, "payload": payload, "force": force})
            # response.text --> "OK"
            return AttrDict({"text": "OK", "raise_for_status": lambda: None})

        def read_user(uuid):
            return self.mo_values

        def update_user(uuid, data):
            payload = {"type": "employee", "uuid": uuid, "data": data}
            return _mo_post("details/edit", payload)

        return AttrDict(
            {
                "read_organisation": lambda: "org_uuid",
                "read_classes_in_facet": lambda x: [
                    [
                        {"uuid": "address_visibility_public_uuid"},
                        {"uuid": "address_visibility_internal_uuid"},
                        {"uuid": "address_visibility_secret_uuid"},
                    ]
                ],
                "read_user": read_user,
                "read_all_users": lambda: [self.mo_values],
                "read_user_engagement": get_e_details("engagement"),
                "get_e_addresses": get_e_details("address"),
                "get_e_itsystems": get_e_details("it"),
                "update_user": update_user,
                "_mo_post": _mo_post,
            }
        )

    def _setup_ad_reader_and_cache_all(self, index, cache_all=True):
        def read_user(cpr, cache_only):
            # We only support one person in our mocking
            if cpr != self.mo_values["cpr"]:
                raise NotImplementedError("Outside mocking")
            # If we got that one person, return it
            return self.ad_values

        def get_settings():
            return self.settings["integrations.ad"][0]

        ad_reader = AttrDict(
            {
                "_get_setting": get_settings,
                "read_user": read_user,
            }
        )
        return ad_reader


class TestADMoSyncMixin(TestADMixin):
    def _initialize_configuration(self):
        def ident(x, *args, **kwargs):
            return x

        self.settings = self._prepare_settings(ident)
        self.mo_values_func = partial(self._prepare_mo_values, ident)
        self.ad_values_func = partial(self._prepare_get_from_ad, ident)
        self.mo_e_username_func = lambda: ""
        self.mo_seed_func = lambda: {}

    def _setup_admosync(
        self,
        transform_settings=None,
        transform_mo_values=None,
        transform_ad_values=None,
        seed_e_username=None,
        seed_mo=None,
    ):
        if transform_settings:
            self.settings = self._prepare_settings(transform_settings)
        if transform_mo_values:
            self.mo_values_func = partial(self._prepare_mo_values, transform_mo_values)
        if transform_ad_values:
            self.ad_values_func = partial(
                self._prepare_get_from_ad, transform_ad_values
            )
        if seed_e_username:
            self.mo_e_username_func = seed_e_username
        if seed_mo:
            self.mo_seed_func = seed_mo

        self.ad_sync = AdMoSyncTestSubclass(
            all_settings=self.settings,
            mo_values_func=self.mo_values_func,
            ad_values_func=self.ad_values_func,
            mo_e_username_func=self.mo_e_username_func,
            mo_seed_func=self.mo_seed_func,
        )



import csv
from datetime import datetime
from typing import List
from unittest import mock
from uuid import UUID

import los_files
import los_org
from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st

from .helpers import HelperMixin
from .helpers import mock_config
from .strategies import csv_buf_from_model
from .strategies import text_except


class TestConsolidatePayloads:
    def test_consolidate_identical(self):
        """Should correctly consolidate identical consecutive payloads"""
        payloads = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2010-12-31"}},
            {"val": 123, "validity": {"from": "2011-01-01", "to": "2011-12-31"}},
            {"val": 123, "validity": {"from": "2012-01-01", "to": "2012-12-31"}},
            {"val": 123, "validity": {"from": "2013-01-01", "to": "2013-12-31"}},
            {"val": 123, "validity": {"from": "2014-01-01", "to": "2014-12-31"}},
        ]

        expected = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2014-12-31"}}
        ]

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_non_identical(self):
        """Should handle two consecutive non-identical payloads"""
        payloads = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2010-12-31"}},
            {"val": 456, "validity": {"from": "2011-01-01", "to": "2011-12-31"}},
        ]

        expected = payloads

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_non_consecutive(self):
        """Should handle non-consecutive payloads"""
        payloads = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2010-12-31"}},
            {"val": 123, "validity": {"from": "2012-01-01", "to": "2012-12-31"}},
        ]

        expected = payloads

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_mixed(self):
        """Should handle a mix between identical and non-identical payloads"""
        payloads = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2010-12-31"}},
            {"val": 123, "validity": {"from": "2011-01-01", "to": "2011-12-31"}},
            {"val": 456, "validity": {"from": "2012-01-01", "to": "2012-12-31"}},
            {"val": 789, "validity": {"from": "2013-01-01", "to": "2013-12-31"}},
            {"val": 789, "validity": {"from": "2014-01-01", "to": "2014-12-31"}},
        ]

        expected = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2011-12-31"}},
            {"val": 456, "validity": {"from": "2012-01-01", "to": "2012-12-31"}},
            {"val": 789, "validity": {"from": "2013-01-01", "to": "2014-12-31"}},
        ]

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_differing(self):
        """Should handle cases where payloads have differing keys"""
        payloads = [
            {"val": 123, "validity": {"from": "2010-01-01", "to": "2010-12-31"}},
            {
                "val": 123,
                "val2": 456,
                "validity": {"from": "2011-01-01", "to": "2011-12-31"},
            },
        ]

        expected = payloads

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_single_element(self):
        """Should return same single payload"""
        payloads = [
            {"val": 456, "validity": {"from": "2011-01-01", "to": "2011-12-31"}}
        ]

        expected = payloads

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual

    def test_consolidate_empty(self):
        """Should trivially handle empty input"""
        payloads = []

        expected = payloads

        actual = los_org.OrgUnitImporter.consolidate_payloads(payloads)

        assert expected == actual


class TestParseOrgUnitCSV:
    @given(csv_buf_from_model(model=los_org.OrgUnit))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_org.OrgUnit)


class TestHandleInitial(HelperMixin):
    @settings(deadline=None)
    @given(st.builds(los_org.OrgUnit), st.uuids())
    def test_handle_initial(self, instance: los_org.OrgUnit, dar_uuid: UUID):
        importer = los_org.OrgUnitImporter()
        with self._mock_dar_lookup((instance.post_address, dar_uuid)):
            with self._mock_read_csv(instance):
                with self._mock_get_client_session():
                    with self._mock_create_details() as mock_create_details:
                        self._run_until_complete(
                            importer.handle_initial("unused_filename.csv")
                        )
                        mock_create_details.assert_called_once()
                        heads = list(mock_create_details.call_args.args[1])
                        assert heads == [
                            {
                                "type": "org_unit",
                                "uuid": str(instance.org_uuid),
                                "user_key": instance.bvn,
                                "parent": {"uuid": str(instance.parent_uuid)},
                                "validity": {
                                    "from": str(instance.start_date),
                                    "to": str(instance.end_date),
                                },
                            }
                        ]


class TestHandleCreate(HelperMixin):
    @given(st.builds(los_org.OrgUnit), st.uuids())
    def test_handle_create(self, instance: los_org.OrgUnit, dar_uuid: UUID):
        importer = los_org.OrgUnitImporter()
        with self._mock_dar_lookup((instance.post_address, dar_uuid)):
            with self._mock_read_csv(instance):
                with self._mock_get_client_session():
                    with self._mock_create_details() as mock_create_details:
                        self._run_until_complete(
                            importer.handle_create("unused_filename.csv")
                        )
                        mock_create_details.assert_called_once()


class TestHandleEdit(HelperMixin):
    @given(st.builds(los_org.OrgUnit), st.uuids(), st.datetimes())
    def test_handle_edit(
        self,
        instance: los_org.OrgUnit,
        dar_uuid: UUID,
        filedate: datetime,
    ):
        importer = los_org.OrgUnitImporter()
        with self._mock_dar_lookup((instance.post_address, dar_uuid)):
            with self._mock_read_csv(instance):
                with self._mock_get_client_session():
                    with self._mock_create_details() as mock_create_details:
                        with self._mock_edit_details() as mock_edit_details:
                            with self._mock_lookup_organisationfunktion():
                                self._run_until_complete(
                                    importer.handle_edit(
                                        "unused_filename.csv", filedate
                                    )
                                )
                                mock_create_details.assert_called_once()
                                mock_edit_details.assert_called_once()


class TestHandleAddresses(HelperMixin):
    @given(st.builds(los_org.OrgUnit))
    def test_finds_failed_addresses(self, instance: los_org.OrgUnit):
        importer = los_org.OrgUnitImporter()
        expected_filename = "filename.csv"

        # Patch DAR lookup function to indicate a failed address lookup for any
        # address passed to it.
        with self._mock_dar_lookup((instance.post_address, None)):
            # Patch `write_failed_addresses` so we can test its arguments
            with mock.patch.object(importer, "write_failed_addresses") as mock_write:
                # Run method under test
                self._run_until_complete(
                    importer.handle_addresses([instance], "filename.csv")
                )
                # Assert args to mocked `write_failed_addresses`
                expected_failed = [
                    los_org.FailedDARLookup(
                        org_uuid=instance.org_uuid,
                        post_address=instance.post_address,
                    )
                ]
                mock_write.assert_called_once_with(expected_failed, expected_filename)

    @given(st.lists(st.builds(los_org.OrgUnit, post_address=st.none())))
    def test_does_nothing_on_empty_addresses(self, org_units: List[los_org.OrgUnit]):
        importer = los_org.OrgUnitImporter()
        with self._mock_dar_lookup(("", None)):
            with mock.patch.object(importer, "write_failed_addresses"):
                self._run_until_complete(
                    importer.handle_addresses(org_units, "filename.csv")
                )
                # Assert that we returned early because we only saw empty
                # addresses in `org_units`.
                assert importer.dar_cache == {}


class TestWriteFailedAddresses:
    # Construct a `FailedDARLookup` where the `post_address` does not contain
    # ASCII NUL characters nor the CSV delimiter character
    # (which in this case is "#" as per customer spec, see
    # `OrgUnitImporter.write_failed_addresses`.)
    @settings(max_examples=1000, deadline=None)
    @given(st.builds(los_org.FailedDARLookup, post_address=text_except(chr(0), "#")))
    def test_writes_csv_to_ftp(self, instance: los_org.FailedDARLookup):
        # Patch `config.get_config` so `write_failed_addresses` sees
        # `MockConfig.queries_dir` instead of actual setting.
        with mock_config(
            queries_dir="/tmp/", import_csv_folder=None, ftp_ssh_key_path=None
        ):
            # Patch `FTPFileSet.write_file` so we can test its arguments
            with mock.patch.object(los_files.FTPFileSet, "write_file") as mock_write:
                # Run `write_failed_addresses`
                importer = los_org.OrgUnitImporter()
                importer.write_failed_addresses([instance], "filename.csv")
                # Assert our mocked `write_file` got the expected args
                mock_write.assert_called_once()
                # First arg is filename, second arg is CSV file (as bytes)
                filename, csv_buf = mock_write.call_args_list[0].args
                # Filename is appended to output filename
                assert filename == "failed_addr_filename.csv"
                # Assert contents of the CSV file built
                csv_rows = list(csv.DictReader(csv_buf, delimiter="#"))
                assert len(csv_rows) == 1
                assert csv_rows[0] == {
                    "org_uuid": str(instance.org_uuid),
                    "post_address": instance.post_address,
                }



# SPDX-FileCopyrightText: 2023 Magenta ApS <https://magenta.dk>
# SPDX-License-Identifier: MPL-2.0
def _virkning(dato="1910-01-01 00:00:00"):
    virkning = {
        "from": dato,
        "to": "infinity",
        "aktoerref": "ddc99abd-c1b0-48c2-aef7-74fea841adae",
        "aktoertypekode": "Bruger",
    }
    return virkning


def lora_facet(bvn, org="ddc99abd-c1b0-48c2-aef7-74fea841adae"):
    attributter = {
        "facetegenskaber": [
            {
                "brugervendtnoegle": bvn,
                "virkning": _virkning(),
            }
        ]
    }
    tilstande = {
        "facetpubliceret": [{"publiceret": "Publiceret", "virkning": _virkning()}]
    }
    relationer = {
        "ansvarlig": [
            {"objekttype": "organisation", "uuid": org, "virkning": _virkning()}
        ]
    }
    facet = {
        "attributter": attributter,
        "tilstande": tilstande,
        "relationer": relationer,
    }
    return facet


def lora_klasse(
    brugervendtnoegle,
    beskrivelse,
    titel,
    dato,
    facet,
    ansvarlig,
    omfang=None,
    overklasse=None,
):
    attributter = {
        "klasseegenskaber": [
            {
                "brugervendtnoegle": brugervendtnoegle,
                "beskrivelse": beskrivelse,
                "titel": titel,
                "virkning": _virkning(dato),
            }
        ]
    }
    tilstande = {
        "klassepubliceret": [{"publiceret": "Publiceret", "virkning": _virkning(dato)}]
    }
    relationer = {
        "facet": [{"uuid": facet, "virkning": _virkning(dato), "objekttype": "Facet"}],
        "overordnetklasse": [{"virkning": _virkning(dato), "objekttype": "Klasse"}],
        "ansvarlig": [
            {
                "uuid": ansvarlig,
                "virkning": _virkning(dato),
                "objekttype": "Organisation",
            }
        ],
    }
    klasse = {
        "attributter": attributter,
        "tilstande": tilstande,
        "relationer": relationer,
    }
    if overklasse is not None:
        klasse["relationer"]["overordnetklasse"][0]["uuid"] = overklasse
    else:
        del klasse["relationer"]["overordnetklasse"]

    if omfang:
        attributter["klasseegenskaber"][0]["omfang"] = omfang

    return klasse



from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import DeclarativeMeta

Base: DeclarativeMeta = declarative_base()


class Compare:
    def __eq__(self, __value: object) -> bool:
        """When comparing objects we need to disregard the fields:

        * ID (as it is auto generated by sql-alchemy)
        * _sa_instance_state
        """

        left = dict(self.__dict__)
        left.pop("id", None)
        left.pop("_sa_instance_state", None)
        right = dict(__value.__dict__)
        right.pop("id", None)
        right.pop("_sa_instance_state", None)
        return left == right


class BaseFacet(Compare):
    bvn = Column(String(250), nullable=False)
    uuid = Column(String(36), nullable=False, primary_key=True)


class WFacet(Base, BaseFacet):
    __tablename__ = "wfacetter"


class Facet(Base, BaseFacet):
    __tablename__ = "facetter"


class BaseKlasse(Compare):
    uuid = Column(String(36), nullable=False, primary_key=True)
    bvn = Column(String(250), nullable=False)
    titel = Column(String(250), nullable=False)
    facet_uuid = Column(String(36))  # , ForeignKey('facetter.uuid'))
    facet_bvn = Column(String(250), nullable=False)


class WKlasse(Base, BaseKlasse):
    __tablename__ = "wklasser"


class Klasse(Base, BaseKlasse):
    __tablename__ = "klasser"


class BaseBruger(Compare):
    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    fornavn = Column(String(250))
    efternavn = Column(String(250))
    kaldenavn_fornavn = Column(String(250))
    kaldenavn_efternavn = Column(String(250))
    cpr = Column(String(250), nullable=True)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WBruger(Base, BaseBruger):
    __tablename__ = "wbrugere"


class Bruger(Base, BaseBruger):
    __tablename__ = "brugere"


class BaseEnhed(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    navn = Column(String(250), nullable=False)
    bvn = Column(String(250), nullable=False)
    forældreenhed_uuid = Column(String(36), primary_key=False)
    enhedstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    enhedstype_titel = Column(String(250), nullable=False)
    enhedsniveau_uuid = Column(String(36))  # ForeignKey('klasser.uuid'))
    enhedsniveau_titel = Column(String(250))
    tidsregistrering_uuid = Column(String(36))  # ForeignKey('klasser.uuid'))
    tidsregistrering_titel = Column(String(250))
    organisatorisk_sti = Column(String(1000))
    leder_uuid = Column(String(36))
    fungerende_leder_uuid = Column(String(36))
    opmærkning_uuid = Column(String(36))  # Export UUID of "org_unit_hierarchy"
    opmærkning_titel = Column(String(250))  # Export title of "org_unit_hierarchy"
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WEnhed(Base, BaseEnhed):
    __tablename__ = "wenheder"


class Enhed(Base, BaseEnhed):
    __tablename__ = "enheder"


class BaseAdresse(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250))
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    værdi = Column(String(250))
    dar_uuid = Column(String(36))  # , ForeignKey('dar_adresser.uuid'))
    adressetype_uuid = Column(String(36))  # , ForeignKey("klasser.uuid"))
    adressetype_bvn = Column(String(250), nullable=False)
    adressetype_scope = Column(String(250), nullable=False)
    adressetype_titel = Column(String(250), nullable=False)
    synlighed_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    synlighed_scope = Column(String(250))
    synlighed_titel = Column(String(250))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WAdresse(Base, BaseAdresse):
    __tablename__ = "wadresser"


class Adresse(Base, BaseAdresse):
    __tablename__ = "adresser"


class BaseEngagement(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    bvn = Column(String(250), nullable=False)
    arbejdstidsfraktion = Column(Integer)
    engagementstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    engagementstype_titel = Column(String(250), nullable=False)
    primærtype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    primærtype_titel = Column(String(250))
    stillingsbetegnelse_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    stillingsbetegnelse_titel = Column(String(250), nullable=False)
    primær_boolean = Column(Boolean)
    udvidelse_1 = Column(String(250))
    udvidelse_2 = Column(String(250))
    udvidelse_3 = Column(String(250))
    udvidelse_4 = Column(String(250))
    udvidelse_5 = Column(String(250))
    udvidelse_6 = Column(String(250))
    udvidelse_7 = Column(String(250))
    udvidelse_8 = Column(String(250))
    udvidelse_9 = Column(String(250))
    udvidelse_10 = Column(String(250))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WEngagement(Base, BaseEngagement):
    __tablename__ = "wengagementer"


class Engagement(Base, BaseEngagement):
    __tablename__ = "engagementer"


class BaseTilknytning(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    tilknytningstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    tilknytningstype_titel = Column(String(250))
    stillingsbetegnelse_uuid = Column(String(36))
    stillingsbetegnelse_titel = Column(String(250))
    it_forbindelse_uuid = Column(String(36))
    startdato = Column(String(10))
    slutdato = Column(String(10))
    primær_boolean = Column(Boolean, nullable=True)
    faglig_organisation = Column(String(250), nullable=True)


class WTilknytning(Base, BaseTilknytning):
    __tablename__ = "wtilknytninger"


class Tilknytning(Base, BaseTilknytning):
    __tablename__ = "tilknytninger"


class BaseOrlov(Compare):
    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    orlovstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    orlovstype_titel = Column(String(250), nullable=False)
    engagement_uuid = Column(String(36))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WOrlov(Base, BaseOrlov):
    __tablename__ = "worlover"


class Orlov(Base, BaseOrlov):
    __tablename__ = "orlover"


class BaseItSystem(Compare):

    uuid = Column(String(36), nullable=False, primary_key=True)
    navn = Column(String(250), nullable=False)


class WItSystem(Base, BaseItSystem):
    __tablename__ = "wit_systemer"


class ItSystem(Base, BaseItSystem):
    __tablename__ = "it_systemer"


class BaseItForbindelse(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    it_system_uuid = Column(String(36))  # , ForeignKey('it_systemer.uuid'))
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    brugernavn = Column(String(250))
    primær_boolean = Column(Boolean)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WItForbindelse(Base, BaseItForbindelse):
    __tablename__ = "wit_forbindelser"


class ItForbindelse(Base, BaseItForbindelse):
    __tablename__ = "it_forbindelser"


class BaseLeder(Compare):
    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    ledertype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    ledertype_titel = Column(String(250), nullable=False)
    niveautype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    niveautype_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WLeder(Base, BaseLeder):
    __tablename__ = "wledere"


class Leder(Base, BaseLeder):
    __tablename__ = "ledere"


class BaseLederAnsvar(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    leder_uuid = Column(String(36))  # , ForeignKey('ledere.uuid'))
    lederansvar_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    lederansvar_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WLederAnsvar(Base, BaseLederAnsvar):
    __tablename__ = "wleder_ansvar"


class LederAnsvar(Base, BaseLederAnsvar):
    __tablename__ = "leder_ansvar"


class BaseKLE(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    kle_aspekt_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    kle_aspekt_titel = Column(String(250), nullable=False)
    kle_nummer_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    kle_nummer_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WKLE(Base, BaseKLE):
    __tablename__ = "wkle"


class KLE(Base, BaseKLE):
    __tablename__ = "kle"


class Kvittering(Base):
    __tablename__ = "kvittering"

    id = Column(Integer, nullable=False, primary_key=True)
    query_tid = Column(DateTime)
    start_levering_tid = Column(DateTime)
    slut_levering_tid = Column(DateTime)


class BaseEnhedssammenkobling(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    enhed1_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    enhed2_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class WEnhedssammenkobling(Base, BaseEnhedssammenkobling):
    __tablename__ = "wenhedssammenkobling"


class Enhedssammenkobling(Base, BaseEnhedssammenkobling):
    __tablename__ = "enhedssammenkobling"


class BaseDARAdresse(Compare):

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    vejkode = Column(String(8))
    vejnavn = Column(String(250))
    husnr = Column(String(8))
    etage = Column(String(16))
    dør = Column(String(8))
    postnr = Column(String(8))
    postnrnavn = Column(String(250))
    kommunekode = Column(String(8))
    adgangsadresseid = Column(String(36))
    betegnelse = Column(String(250))


class WDARAdresse(Base, BaseDARAdresse):
    __tablename__ = "wdar_adresser"


class DARAdresse(Base, BaseDARAdresse):
    __tablename__ = "dar_adresser"


sql_type = (
    Facet
    | Klasse
    | Bruger
    | Enhed
    | Adresse
    | Engagement
    | Tilknytning
    | Orlov
    | ItSystem
    | ItForbindelse
    | Leder
    | LederAnsvar
    | KLE
    | Enhedssammenkobling
    | DARAdresse
)



from typing import List
from typing import Set
from unittest.mock import patch

import pytest
from hypothesis import given
from hypothesis import strategies as st
from parameterized import parameterized

from ..ad_jinja_filters import name_to_email_address
from .mocks import MO_UUID
from .test_ad_writer import _TestRealADWriter


class TestFirstAddressOfTypeIntegration(_TestRealADWriter):
    _template_to_ad_fields = {
        "someAdField": "{{ mo_values['employee_addresses']|first_address_of_type('address-type-uuid') }}"
    }
    _expected_fragment = '"someAdField"="address-value"'

    def test_jinja_env_has_template_filter_first_address_of_type(self):
        ad_writer = self._prepare_adwriter()
        self.assertIn("first_address_of_type", ad_writer._environment.filters)

    def test_first_address_of_type_in_create_user_command(self):
        ad_writer = self._prepare_adwriter(
            template_to_ad_fields=self._template_to_ad_fields
        )
        create_cmd, _ = ad_writer._preview_create_command(MO_UUID)
        self.assertIn(self._expected_fragment, create_cmd)

    def test_first_address_of_type_in_sync_user_command(self):
        ad_writer = self._prepare_adwriter(
            template_to_ad_fields=self._template_to_ad_fields
        )
        sync_cmd, _, _, _ = ad_writer._preview_sync_command(MO_UUID, "user_sam")
        self.assertIn(self._expected_fragment, sync_cmd)


class TestNameToEmailAddressIntegration(_TestRealADWriter):
    _template_to_ad_fields = {
        "someAdField": "{{ mo_values['full_name']|name_to_email_address }}"
    }

    @parameterized.expand(
        [
            (
                # No taken emails, use variant A.
                set(),
                "tester.testesen@upn_end",
            ),
            (
                # Variant A taken, use variant B.
                {"tester.testesen@upn_end"},
                "t.testesen@upn_end",
            ),
            (
                # Variants A and B taken, use variant C.
                {"tester.testesen@upn_end", "t.testesen@upn_end"},
                "tester.t@upn_end",
            ),
            (
                # Variants A, B and C taken, use variant D.
                # Start serial number at 3 since no other D variants are taken.
                {"tester.testesen@upn_end", "t.testesen@upn_end", "tester.t@upn_end"},
                "tester.testesen.3@upn_end",
            ),
            (
                # Variants A, B, C and D taken, use variant D with next numeric suffix.
                {
                    "tester.testesen@upn_end",
                    "t.testesen@upn_end",
                    "tester.t@upn_end",
                    "tester.testesen.3@upn_end",
                    "irrelevant.4@upn_end",
                },
                "tester.testesen.4@upn_end",
            ),
        ]
    )
    def test_name_to_email_address_in_create_user_command(
        self, taken_emails: Set[str], expected_result: str
    ):
        with patch(
            "integrations.ad_integration.ad_reader.ADParameterReader.__init__",
            return_value=None,
        ):
            with patch(
                "integrations.ad_integration.ad_reader.ADParameterReader"
                ".get_all_email_values",
                return_value=taken_emails,
            ):
                ad_writer = self._prepare_adwriter(
                    template_to_ad_fields=self._template_to_ad_fields,
                )
                create_cmd, _ = ad_writer._preview_create_command(MO_UUID)

        self.assertIn(f'"someAdField"="{expected_result}"', create_cmd)


class TestNameToEmailAddress:
    _letters = st.characters(whitelist_categories=("Ll", "Lu"), max_codepoint=500)

    @given(
        st.one_of(
            # Too short input - not enough name parts
            st.lists(st.text(min_size=1, alphabet=_letters), min_size=0),
            # Sufficient input - two or more name parts
            st.lists(st.text(min_size=2, alphabet=_letters), min_size=1),
        ),
        # Separator between name parts
        st.one_of(st.just(" "), st.just("-")),
    )
    def test_input(self, name_parts: List[str], separator: str):
        name = separator.join(name_parts)
        ctx = {
            "_upn_end": "upn_end",
            "_get_all_ad_emails": lambda: set(),
        }
        if name.count(" ") or name.count("-"):
            # Name contains at least one space or dash. This means we can extract at
            # least two "name parts."
            email = name_to_email_address(ctx, name)
            assert email.endswith("@upn_end")
            value = email.split("@upn_end")[0]
            assert len(value) > 0
            assert value.count("@") == 0
            assert all(ch.isalnum() or ch == "." for ch in value)
        else:
            # Name contains fewer than two name parts, and we cannot generate an email.
            with pytest.raises(ValueError):
                name_to_email_address(ctx, name)



from abc import ABC

import click
import pandas as pd
import xlsxwriter.worksheet
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup

from integrations.kle.kle_import_export import ASPECT_MAP
from integrations.kle.kle_import_export import Aspects
from integrations.kle.kle_import_export import KLEAnnotationIntegration


class KLEXLSXIntegration(KLEAnnotationIntegration, ABC):
    def __init__(self):
        super().__init__()

        self.xlsx_file = self.settings.integrations_kle_xlsx_file_path


class KLEXLSXExporter(KLEXLSXIntegration):
    """Export KLE annotation as CSV files bundled in a spreadsheet."""

    def __init__(self, multiple_responsible=False):
        self.multiple_responsible = multiple_responsible
        super().__init__()

    @staticmethod
    def write_rows(worksheet: xlsxwriter.worksheet.Worksheet, data: list):

        for index, row in enumerate(data):
            worksheet.write_row(index, 0, row)

    @staticmethod
    def get_org_unit_validation(column: str):
        return (
            "{0}1:{0}1048576".format(column),
            {"validate": "list", "source": "=Org!$B$2:$B$1048576"},
        )

    @staticmethod
    def get_kle_validation(column: str):
        return (
            "{0}1:{0}1048576".format(column),
            {"validate": "list", "source": "=KLE!$C$2:$C$1048576"},
        )

    @staticmethod
    def get_column_width(data, field: str):
        field_lengths = [len(row[field]) for row in data]
        return max(field_lengths)

    def add_org_unit_sheet(self, workbook, org_units):
        worksheet = workbook.add_worksheet(name="Org")

        rows = [(org_unit["uuid"], org_unit["combined"]) for org_unit in org_units]

        worksheet.set_column(0, 0, width=self.get_column_width(org_units, "uuid"))
        worksheet.set_column(1, 1, width=self.get_column_width(org_units, "combined"))

        rows.insert(0, ("UUID", "Navn"))

        self.write_rows(worksheet, rows)

    def add_kle_sheet(self, workbook: xlsxwriter.Workbook, kle_numbers: list):
        worksheet = workbook.add_worksheet(name="KLE")

        rows = [(kle["uuid"], kle["user_key"], kle["name"]) for kle in kle_numbers]

        rows.insert(0, ("UUID", "EmneNr", "EmneTitel"))

        worksheet.set_column(0, 0, width=self.get_column_width(kle_numbers, "uuid"))
        worksheet.set_column(1, 1, width=self.get_column_width(kle_numbers, "user_key"))
        worksheet.set_column(2, 2, width=self.get_column_width(kle_numbers, "name"))

        self.write_rows(worksheet, rows)

    def add_ansvarlig_sheet(self, workbook, kle_numbers, org_units):
        worksheet = workbook.add_worksheet(name="Ansvarlig")

        def calculate_level(kle_number: str):
            """
            We calculate the level, by how many dots are in the key
            E.g. 00 is 1, 00.01 is 2, 00.01.32 is 3
            """
            return str(kle_number.count(".") + 1)

        rows = [(kle["level"], kle["user_key"], kle["name"], "") for kle in kle_numbers]
        rows.insert(0, ("Niveau", "EmneNr", "EmneTitel", "EnhedNavn"))

        worksheet.data_validation(*self.get_org_unit_validation(column="D"))

        worksheet.set_column(1, 1, width=self.get_column_width(kle_numbers, "user_key"))
        worksheet.set_column(2, 2, width=self.get_column_width(kle_numbers, "name"))
        worksheet.set_column(3, 3, width=self.get_column_width(org_units, "combined"))

        self.write_rows(worksheet, rows)

    def add_kle_relation_sheet(self, sheet_name, workbook, kle_numbers, org_units):
        worksheet = workbook.add_worksheet(name=sheet_name)

        rows = [("EnhedNavn", "KLE")]

        worksheet.data_validation(*self.get_org_unit_validation("A"))
        worksheet.data_validation(*self.get_kle_validation("B"))

        worksheet.set_column(0, 0, width=self.get_column_width(org_units, "combined"))
        worksheet.set_column(1, 1, width=self.get_column_width(kle_numbers, "name"))

        self.write_rows(worksheet, rows)

    @staticmethod
    def convert_org_units(org_units):
        return [
            {
                "combined": "{} - {}".format(unit["name"], unit["uuid"]),
                **unit,
            }
            for unit in org_units
        ]

    @staticmethod
    def convert_kle_numbers(kle_numbers):
        def calculate_level(kle_number: str):
            """
            We calculate the level, by how many dots are in the key
            E.g. 00 is 1, 00.01 is 2, 00.01.32 is 3
            """
            return str(kle_number.count(".") + 1)

        return [
            {
                "level": calculate_level(kle["user_key"]),
                **kle,
            }
            for kle in kle_numbers
        ]

    def run(self):
        workbook = xlsxwriter.Workbook(self.xlsx_file)

        org_units = sorted(
            self.convert_org_units(self.get_all_org_units_from_mo()),
            key=lambda x: x["name"],
        )
        kle = sorted(
            self.convert_kle_numbers(self.get_kle_classes_from_mo()),
            key=lambda x: x["user_key"],
        )

        self.add_org_unit_sheet(workbook, org_units)
        self.add_kle_sheet(workbook, kle)

        if self.multiple_responsible:
            self.add_kle_relation_sheet("Ansvarlig", workbook, kle, org_units)
        else:
            self.add_ansvarlig_sheet(workbook, kle, org_units)

        self.add_kle_relation_sheet("Indsigt", workbook, kle, org_units)
        self.add_kle_relation_sheet("Udførende", workbook, kle, org_units)

        # Bold column headers for all sheets
        bold = workbook.add_format({"bold": 1})
        for sheet in workbook.worksheets():
            sheet.set_row(0, cell_format=bold)

        workbook.close()


class KLEXLSXImporter(KLEXLSXIntegration):
    def handle_sheet(
        self,
        sheet,
        org_unit_field,
        kle_field,
        data_map,
        org_unit_map,
        kle_map,
        kle_aspect,
    ):
        for row in sheet.iterrows():
            index, data = row

            kle_uuid = kle_map.get(data.__getattr__(kle_field))
            org_unit_uuid = org_unit_map.get(data.__getattr__(org_unit_field))

            if not kle_uuid or not org_unit_uuid:
                continue

            org_unit = data_map.setdefault(org_unit_uuid, {})
            org_unit.setdefault(kle_uuid, set()).add(kle_aspect)

    def run(self):

        # Read sheet
        xlsx_file = pd.ExcelFile(self.xlsx_file)

        sheets = {
            sheet_name: xlsx_file.parse(sheet_name)
            for sheet_name in xlsx_file.sheet_names
        }

        org_unit_map = {row.Navn: row.UUID for index, row in sheets["Org"].iterrows()}
        kle_map = {row.EmneTitel: row.UUID for index, row in sheets["KLE"].iterrows()}

        data_map = {}

        self.handle_sheet(
            sheets["Ansvarlig"],
            org_unit_field="EnhedNavn",
            kle_field="KLE",
            data_map=data_map,
            org_unit_map=org_unit_map,
            kle_map=kle_map,
            kle_aspect=Aspects.Ansvarlig,
        )
        self.handle_sheet(
            sheets["Indsigt"],
            org_unit_field="EnhedNavn",
            kle_field="KLE",
            data_map=data_map,
            org_unit_map=org_unit_map,
            kle_map=kle_map,
            kle_aspect=Aspects.Indsigt,
        )
        self.handle_sheet(
            sheets["Udførende"],
            org_unit_field="EnhedNavn",
            kle_field="KLE",
            data_map=data_map,
            org_unit_map=org_unit_map,
            kle_map=kle_map,
            kle_aspect=Aspects.Udfoerende,
        )

        payloads = self.generate_payloads(data_map)

        self.post_payloads_to_mo(payloads)

    def generate_payloads(self, data):
        aspect_classes = self.get_aspect_classes_from_mo()
        aspect_map = {
            ASPECT_MAP[clazz["scope"]]: clazz["uuid"] for clazz in aspect_classes
        }

        payloads = []
        for unit, info in data.items():
            for kle_uuid, aspects in info.items():

                aspect_uuids = [aspect_map[aspect] for aspect in aspects]

                payload = {
                    "type": "kle",
                    "org_unit": {"uuid": unit},
                    "kle_aspect": [{"uuid": uuid} for uuid in aspect_uuids],
                    "kle_number": {"uuid": kle_uuid},
                    "validity": {"from": "1920-01-01", "to": None},
                }
                payloads.append(payload)

        return payloads


@click.command()
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--import", is_flag=True)
@optgroup.option("--export", is_flag=True)
@click.option("--multiple-responsible", is_flag=True)
def cli(**args):
    if args["import"]:
        importer = KLEXLSXImporter()
        importer.run()

    if args["export"]:
        exporter = KLEXLSXExporter(args["multiple_responsible"])
        exporter.run()


if __name__ == "__main__":
    cli()
